from openmdao.api import ExplicitComponent, Group, Problem
import numpy as np

class PlantFinance(ExplicitComponent):
    def initialize(self):
        self.options.declare('verbosity',default=False)
        
    def setup(self):

        # Inputs
        self.add_input('machine_rating',    val=0.0, units='kW',        desc='Rating of the turbine')
        self.add_input('tcc_per_kW' ,       val=0.0, units='USD/kW',    desc='A wind turbine capital cost')
        self.add_discrete_input('turbine_number',    val=0,             desc='Number of turbines at plant')
        self.add_input('bos_per_kW',        val=0.0, units='USD/kW',    desc='Balance of system costs of the turbine')
        self.add_input('opex_per_kW',       val=0.0, units='USD/kW/yr', desc='Average annual operational expenditures of the turbine')
        self.add_input('park_aep',          val=0.0, units='kW*h',      desc='Annual Energy Production of the wind plant')
        self.add_input('turbine_aep',       val=0.0, units='kW*h',      desc='Annual Energy Production of the wind turbine')

        # parameters
        self.add_input('wake_loss_factor',  val=0.15,                   desc='The losses in AEP due to waked conditions')
        self.add_input('fixed_charge_rate', val=0.075,                  desc = 'Fixed charge rate for coe calculation')

        #Outputs
        self.add_output('lcoe',             val=0.0, units='USD/kW/h',  desc='Levelized cost of energy for the wind plant')

        self.declare_partials('*','*')
        
    
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack parameters
        t_rating    = inputs['machine_rating']
        n_turbine   = discrete_inputs['turbine_number']
        tcc_per_kW  = inputs['tcc_per_kW'] 
        bos_per_kW  = inputs['bos_per_kW'] 
        opex_per_kW = inputs['opex_per_kW'] 
        fcr         = inputs['fixed_charge_rate']
        wlf         = inputs['wake_loss_factor']
        turb_aep    = inputs['turbine_aep']
        c_turbine       = tcc_per_kW * t_rating
        c_bos_turbine   = bos_per_kW * t_rating
        c_opex_turbine  = opex_per_kW * t_rating
        
        
        # Run a few checks on the inputs
        if n_turbine == 0:
            exit('ERROR: The number of the turbines in the plant is not initialized correctly and it is currently equal to 0. Check the connections to Plant_FinanceSE')
        
        if c_turbine == 0:
            exit('ERROR: The cost of the turbines in the plant is not initialized correctly and it is currently equal to 0 USD. Check the connections to Plant_FinanceSE')
            
        if c_bos_turbine == 0:
            print('WARNING: The BoS costs of the turbine are not initialized correctly and they are currently equal to 0 USD. Check the connections to Plant_FinanceSE')
        
        if c_opex_turbine == 0:
            print('WARNING: The Opex costs of the turbine are not initialized correctly and they are currently equal to 0 USD. Check the connections to Plant_FinanceSE')
        
        if inputs['park_aep'] == 0:
            if turb_aep != 0:
                park_aep     =  n_turbine * turb_aep * (1. - wlf)
                dpark_dtaep  =  n_turbine            * (1. - wlf)
                dpark_dnturb =              turb_aep * (1. - wlf)
                dpark_dwlf   = -n_turbine * turb_aep
                dpark_dpaep  = 0.0
            else:
                exit('ERROR: AEP is not connected properly. Both turbine_aep and park_aep are currently equal to 0 Wh. Check the connections to Plant_FinanceSE')
        else:
            park_aep    = inputs['park_aep']
            dpark_dpaep = 1.0
            dpark_dtaep = dpark_dnturb = dpark_dwlf = 0.0
        
        npr           = n_turbine * t_rating # net park rating, used in net energy capture calculation below
        dnpr_dnturb   =             t_rating
        dnpr_dtrating = n_turbine
        
        nec           = park_aep     / npr # net energy rating, per COE report
        dnec_dwlf     = dpark_dwlf   / npr
        dnec_dtaep    = dpark_dtaep  / npr
        dnec_dpaep    = dpark_dpaep  / npr
        dnec_dnturb   = dpark_dnturb / npr - dnpr_dnturb   * nec / npr
        dnec_dtrating =                    - dnpr_dtrating * nec / npr
        
        icc     = (c_turbine + c_bos_turbine) / t_rating #$/kW, changed per COE report
        c_opex  = (c_opex_turbine) / t_rating  # $/kW, changed per COE report

        dicc_dtrating   = -icc / t_rating
        dcopex_dtrating = -c_opex / t_rating
        dicc_dcturb = dicc_dcbos = dcopex_dcopex = 1.0 / t_rating
           
        #compute COE and LCOE values
        lcoe = ((icc * fcr + c_opex) / nec) # changed per COE report
        outputs['lcoe'] = lcoe

        self.J = {}
        self.J['lcoe', 'tcc_per_kW'       ] = dicc_dcturb*fcr /nec
        self.J['lcoe', 'turbine_number'   ] = - dnec_dnturb*lcoe/nec
        self.J['lcoe', 'bos_per_kW'       ] = dicc_dcbos *fcr /nec
        self.J['lcoe', 'opex_per_kW'      ] = dcopex_dcopex   /nec
        self.J['lcoe', 'fixed_charge_rate'] = icc / nec
        self.J['lcoe', 'wake_loss_factor' ] = -dnec_dwlf *lcoe/nec
        self.J['lcoe', 'turbine_aep'      ] = -dnec_dtaep*lcoe/nec
        self.J['lcoe', 'park_aep'         ] = -dnec_dpaep*lcoe/nec
        self.J['lcoe', 'machine_rating'   ] = (dicc_dtrating*fcr + dcopex_dtrating)/nec - dnec_dtrating*lcoe/nec
        
        if self.options['verbosity']:
            print('################################################')
            print('Computation of LCoE from Plant_FinanceSE')
            print('Number of turbines in the park    %u'              % n_turbine)
            print('Turbine rating                    %.2f kW'         % t_rating)
            print('Turbine capital cost per kW       %.2f USD/kW'     % tcc_per_kW)
            print('BoS costs per kW                  %.2f USD/kW'     % bos_per_kW)
            print('Opex costs per kW                 %.2f USD/kW'     % opex_per_kW)
            print('Fixed charge rate                 %.2f %%'         % (fcr * 100.))
            print('Wake loss factor                  %.2f %%'         % (wlf * 100.))
            print('AEP of the single turbine         %.2f GWh'        % (turb_aep * 1.e-006))
            print('AEP of the wind plant             %.2f GWh'        % (park_aep * 1.e-006))
            print('Initial capital costs per kW      %.2f $/kW'       % icc)
            print('Total initial capital cost        %.2f M USD'      % (icc * n_turbine * t_rating * 1.e-006))  
            print('Opex costs of the park            %.2f M USD/yr'   % (c_opex_turbine * n_turbine * 1.e-006))              
            print('Net energy capture                %.2f MWh/MW/yr'  % nec)
            print('LCoE                              %.2f USD/MWh'    % (lcoe  * 1.e003)) #removed "coe", best to have only one metric for cost
            print('################################################')
            
                    

    def compute_partials(self, inputs, J, discrete_inputs):
        J = self.J



    
class Finance(Group):
    
    def setup(self):
        self.add_subsystem('plantfinancese', PlantFinance(verbosity = True), promotes=['*'])


if __name__ == "__main__":
    # Initialize OpenMDAO problem and FloatingSE Group
    prob = Problem()
    prob.model=Finance() # runs script
    prob.setup()

    prob['machine_rating']          = 2.32 * 1.e+003       # kW
    prob['tcc_per_kW']              = 1093                 # USD/kW
    prob['turbine_number']          = 87.
    prob['opex_per_kW']             = 43.56                # USD/kW/yr Source: 70 $/kW/yr, updated from report, (70 is on the high side)
    prob['fixed_charge_rate']       = 0.079216644          # 7.9 % confirmed from report
    prob['bos_per_kW']              = 517.                 # USD/kW from appendix of report
    prob['wake_loss_factor']        = 0.15                 # confirmed from report 
    prob['turbine_aep']             = 9915.95 * 1.e+003    # confirmed from report 
    
    prob.run_driver()




"""Tests suite for ORBIT."""

__author__ = ["Alicia Key", "Parangat Bhaskar", "Annika Eberle", "Owen Roberts", "Garrett Barter"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Alicia Key"
__email__ = "alicia.key@nrel.gov"


import os, sys
import numpy as np
import multiprocessing as mp
from wisdem.aeroelasticse.runFAST_pywrapper import runFAST_pywrapper_batch
from wisdem.aeroelasticse.Turbsim_mdao.pyturbsim_wrapper import pyTurbsim_wrapper
from wisdem.aeroelasticse.CaseGen_General import CaseGen_General, save_case_matrix_direct

from openmdao.core.mpi_wrap import MPI
if MPI:
    from openmdao.api import PetscImpl as impl
    from mpi4py import MPI
    from petsc4py import PETSc
else:
    from openmdao.api import BasicImpl as impl

class runTS_pywrapper_batch(object):

    def __init__(self, filedict, case_list=None, case_name_list = None, overwrite=True):
        self.case_list          = case_list ### user needs to set this if they don't provide it.
        self.case_name_list = case_name_list ### ditto
        self.filedict = filedict

        self.overwrite = overwrite

    def common_init(self):
        pass
#        if not os.path.exists(self.FAST_runDirectory):
#            os.makedirs(self.FAST_runDirectory)

    def run_serial(self):
        # Run batch serially
        self.common_init()

        res = []
        for case_idx in range(len(self.case_list)):
            case = self.case_list[case_idx]
            case_name = self.case_name_list[case_idx]
            dat = [case,self.filedict, case_idx, case_name, self.overwrite]
            r = tseval(dat)
            res.append(r)
        for r in res:
            idx = r[0]
            fname = r[1]
            self.case_list[idx]['tswind_file'] = fname

    def run_multi(self, cores=None):
        # Run cases in parallel, threaded with multiprocessing module

        self.common_init()

        if not cores:
            cores = mp.cpu_count()
        pool = mp.Pool(cores)

        caseNfile = []
        for case_idx in range(len(self.case_list)):
            case = self.case_list[case_idx]
            case_name = self.case_name_list[case_idx]
            caseNfile.append([case,self.filedict, case_idx, case_name, self.overwrite])
        res = pool.map(tseval, caseNfile)
        # map returns a list of all the return vals from each call.
        # now add them to the cases:
        for r in res:
            idx = r[0]
            fname = r[1]
            self.case_list[idx]['tswind_file'] = fname

    def run_mpi(self, mpi_comm_map_down):

        self.common_init()

        comm = MPI.COMM_WORLD
        # size = comm.Get_size()
        rank = comm.Get_rank()
        sub_ranks = mpi_comm_map_down[rank]
        size = len(sub_ranks)

        caseNfile = []
        for case_idx in range(len(self.case_list)):
            case = self.case_list[case_idx]
            case_name = self.case_name_list[case_idx]
            caseNfile.append([case,self.filedict, case_idx, case_name, self.overwrite])

        N_cases = len(caseNfile)
        N_loops = int(np.ceil(float(N_cases)/float(size)))

        U_out = []
        WindFile_out = []
        WindFile_type_out = []
        for i in range(N_loops):
            idx_s    = i*size
            idx_e    = min((i+1)*size, N_cases)

            for j, var_vals in enumerate(caseNfile[idx_s:idx_e]):
                data   = [tseval, var_vals]
                rank_j = sub_ranks[j]
                comm.send(data, dest=rank_j, tag=0)

            for j, var_vals in enumerate(caseNfile[idx_s:idx_e]):
                rank_j = sub_ranks[j]
                data_out = comm.recv(source=rank_j, tag=1)

                idx   = data_out[0]
                fname = data_out[1]
                self.case_list[idx]['tswind_file'] = fname

    # def run_mpi(self, comm=None):
    #     from mpi4py import MPI

    #     self.common_init()

    #     # file management
    #     if not os.path.exists(self.FAST_runDirectory):
    #         os.makedirs(self.FAST_runDirectory)

    #     # mpi comm management
    #     if not comm:
    #         comm = MPI.COMM_WORLD
    #     size = comm.Get_size()
    #     rank = comm.Get_rank()

    #     N_cases = len(self.case_list)
    #     N_loops = int(np.ceil(float(N_cases)/float(size)))

    #     if rank == 0:
    #         caseNfile = []
    #         for case_idx in range(len(self.case_list)):
    #             case = self.case_list[case_idx]
    #             case_name = self.case_name_list[case_idx]
    #             caseNfile.append([case,self.filedict, case_idx, case_name])
    #     else:
    #         caseNfile = []

    #     output = []
    #     for i in range(N_loops):
    #         # if # of cases left to run is less than comm size, split comm
    #         n_resid = N_cases - i*size
    #         if n_resid < size: 
    #             split_comm = True
    #             color = np.zeros(size)
    #             for i in range(n_resid):
    #                 color[i] = 1
    #             color = [int(j) for j in color]
    #             comm_i  = MPI.COMM_WORLD.Split(color_i, 1)
    #         else:
    #             split_comm = False
    #             comm_i = comm

    #         # position in case list
    #         idx_s  = i*size
    #         idx_e  = min((i+1)*size, N_cases)

    #         # scatter out cases
    #         if split_comm:
    #             if color[rank] == 1:
    #                 case_data_i = comm_i.scatter(caseNfile[idx_s:idx_e], root=0)    
    #         else:
    #             case_data_i = comm_i.scatter(caseNfile[idx_s:idx_e], root=0)
            
    #         # eval
    #         out = tseval(case_data_i)

    #         # gather results
    #         if split_comm:
    #             if color[rank] == 1:
    #                 output_i = comm_i.gather(out, root=0)
    #         else:
    #             output_i = comm_i.gather(out, root=0)

    #         if rank == 0:
    #             output.extend(output_i)

    #     if rank == 0:
    #         output_fname = []
    #         for outi in output:
    #             output_fname.append(outi[1])

    #     return output
        
        



def tseval(dat):
    # Batch FAST pyWrapper call, as a function outside the runFAST_pywrapper_batch class for pickle-ablility
    case = dat[0]
    filedict = dat[1]
    case_idx = dat[2]
    case_name = dat[3]
    # print("running ts", case, filedict)
    pyturb = pyTurbsim_wrapper(filedict, case, case_name) # initialize runner with case variable inputs
    try:
        overwrite = dat[4]
        pyturb.overwrite = overwrite
    except:
        pass

#        pyturb.ny = 20 # example of changing an attribute
    pyturb.execute() # run
    #case['tswind_file'] = pyturb.tswind_file  ### need to really return it!
    #case['tswind_dir'] = pyturb.tswind_dir
    #batch_wrapper.set_case_outputs(case_idx, pyturb.tswind_dir, pyturb.tswind_file) ### future reference: does not work!
    return [case_idx, pyturb.tswind_file]

######################### example drivers.... #####

def example_runTSFAST_Batch():
    fastBatch = runFAST_pywrapper_batch(FAST_ver='OpenFAST')


    mac = False
    if mac:
        fastBatch.FAST_InputFile = '5MW_Land_DLL_WTurb-fast.fst'   # FAST input file (ext=.fst)
        fastBatch.FAST_exe = '/Users/pgraf/opt/openfast/openfast/install/bin/openfast'   # Path to executable
        fastBatch.FAST_directory = '/Users/pgraf/work/wese/templates/openfast/5MW_Land_DLL_WTurb-ModifiedForPyturbsim'
    else:
        fastBatch.FAST_InputFile = '5MW_Land_DLL_WTurb-fast.fst'   # FAST input file (ext=.fst)
        fastBatch.FAST_exe = '/home/pgraf/opt/openfast/openfast/install/bin/openfast'   # Path to executable
        fastBatch.FAST_directory = '/home/pgraf/projects/wese/newaero/templates/openfast/5MW_Land_DLL_WTurb-ModifiedForPyturbsim'

    fastBatch.FAST_runDirectory = 'temp/OpenFAST'
    fastBatch.debug_level = 2

    ## Define case list explicitly
    # case_list = [{}, {}]
    # case_list[0]['Fst', 'TMax'] = 4.
    # case_list[1]['Fst', 'TMax'] = 5.
    # case_name_list = ['test01', 'test02']


    tmaxs = [60.0]
    vhubs = [12.0]
    rhos = [0.0]
    seeds = [1,2]

    case_inputs = {}
    case_inputs[("WrFMTFF")] = {'vals':[False], 'group':0}
    case_inputs[("WrBLFF")] = {'vals':[True], 'group':0}
    case_inputs[("WrADFF")] = {'vals':[True], 'group':0}
    case_inputs[("WrADTWR")] = {'vals':[False], 'group':0}   #WrADTWR
    case_inputs[("TMax")] = {'vals':[t + 30 for t in tmaxs], 'group':0}
    #case_inputs[("TMax")] = {'vals':[t for t in tmaxs], 'group':0}
    case_inputs[("Vhub")] = {'vals':vhubs, 'group':1}
    case_inputs[("Rho")] = {'vals':rhos, 'group':1}
    case_inputs[("RandSeed1")] = {'vals':seeds, 'group':2}
    case_inputs[("RandSeed")] = {'vals':seeds, 'group':2}
    ts_case_list, ts_case_name_list = CaseGen_General(case_inputs, dir_matrix='', namebase='pyTurbsim_testing')

    if mac:
        ts_filedict = {
            'ts_dir':"/Users/pgraf/work/wese/templates/turbsim/pyturbsim/",
            'ts_file':"evans_faster.inp",
            'run_dir':"test_ts_run_dir"}
    else:
        ts_filedict = {
            'ts_dir':"/home/pgraf/projects/wese/newaero/templates/turbsim/pyturbsim/",
            'ts_file':"evans_faster.inp",
            'run_dir':"test_ts_run_dir"}
        

    tsBatch = runTS_pywrapper_batch(ts_filedict, ts_case_list, ts_case_name_list)
    #tsBatch.run_serial()
    tsBatch.run_multi(4)
    ### At this point turbsim is done running and the .bts file names have been added to each case in tsBatch.case_list

    ## Generate case list using General Case Generator
    ## Specify several variables that change independently or collectly
    case_inputs = {}
    case_inputs[("Fst","TMax")] = {'vals':tmaxs, 'group':0}
#    case_inputs[("AeroDyn15","TwrPotent")] = {'vals':[0], 'group':0}
#    case_inputs[("AeroDyn15","TwrShadow")] = {'vals':['False'], 'group':0}
#    case_inputs[("AeroDyn15","TwrAero")] = {'vals':['True'], 'group':0}
    case_inputs[("InflowWind","WindType")] = {'vals':[3], 'group':0}   # 1 = steady, 3 = turbsim binary
    case_inputs[("Fst","OutFileFmt")] = {'vals':[1], 'group':0}
    case_inputs[("InflowWind","HWindSpeed")] = {'vals':vhubs, 'group':1}
    case_inputs[("InflowWind","Rho")] = {'vals':rhos, 'group':1}
    case_inputs[("InflowWind","RandSeed1")] = {'vals':seeds, 'group':2}

    # case_inputs[("ElastoDyn","RotSpeed")] = {'vals':[9.156, 10.296, 11.431, 11.89, 12.1], 'group':1}
    # case_inputs[("ElastoDyn","BlPitch1")] = {'vals':[0., 0., 0., 0., 3.823], 'group':1}
    # case_inputs[("ElastoDyn","BlPitch2")] = case_inputs[("ElastoDyn","BlPitch1")]
    # case_inputs[("ElastoDyn","BlPitch3")] = case_inputs[("ElastoDyn","BlPitch1")]
    # case_inputs[("ElastoDyn","GenDOF")] = {'vals':['True','False'], 'group':2}
    case_list, case_name_list = CaseGen_General(case_inputs, dir_matrix=fastBatch.FAST_runDirectory, namebase='testing')
    # manually adding the wind file names from turb sim run.  This seems a little sketchy
    ### Note to Evan: I feel I should be able to use ONE call to CaseGen_General, instead of one for turbsim, and one for FAStTunr1
    ## thoughts?  solution?
    #####
    print("ADDING WIND FILE NAMES")
    for i in range(len(case_list)):
        case_list[i][("InflowWind","Filename")] = tsBatch.case_list[i]['tswind_file']
        case_list[i][("InflowWind","FilenameRoot")] = tsBatch.case_list[i]['tswind_file'].replace(".wnd", "")
#        case_list[i][("InflowWind","InflowFile")] = tsBatch.case_name_list[i]
        # print(case_list[i])

    fastBatch.case_list = case_list
    fastBatch.case_name_list = case_name_list

    #fastBatch.run_serial()
    fastBatch.run_multi(4)
    ## at this point FAST has run, and the output file names have been added to the case list.
    print("ADDED FAST OUTPUT FILE NAMES")
    for i in range(len(fastBatch.case_list)):
        print(fastBatch.case_list[i])
    save_case_matrix_direct(fastBatch.case_list, dir_matrix=os.path.join(os.getcwd(),fastBatch.FAST_runDirectory))

if __name__=="__main__":


    example_runTSFAST_Batch()


from wisdem.ccblade import CCAirfoil, CCBlade as CCBlade
from openmdao.api import ExplicitComponent
import numpy as np

cosd = lambda x: np.cos(np.deg2rad(x))
sind = lambda x: np.sin(np.deg2rad(x))

class CCBladeGeometry(ExplicitComponent):
    def initialize(self):
        self.options.declare('NINPUT')
        
        
    def setup(self):
        NINPUT = self.options['NINPUT']
        
        self.add_input('Rtip', val=0.0, units='m', desc='tip radius')
        self.add_input('precurve_in', val=np.zeros(NINPUT), desc='prebend distribution', units='m')
        self.add_input('presweep_in', val=np.zeros(NINPUT), desc='presweep distribution', units='m')
        self.add_input('precone', val=0.0, desc='precone angle', units='deg')
        self.add_output('R', val=0.0, units='m', desc='rotor radius')
        self.add_output('diameter', val=0.0, units='m')
        self.add_output('precurveTip', val=0.0, units='m', desc='tip prebend')
        self.add_output('presweepTip', val=0.0, units='m', desc='tip sweep')
        
        #self.declare_partials('R', '*')
        #self.declare_partials('diameter', '*')
        #self.declare_partials('diameter', 'R')
        
    def compute(self, inputs, outputs):
        
        self.Rtip           = inputs['Rtip']
        self.precone        = inputs['precone']
        
        outputs['precurveTip']  = inputs['precurve_in'][-1]
        outputs['presweepTip']  = inputs['presweep_in'][-1]
        self.precurveTip        = outputs['precurveTip']
        self.R = self.Rtip*cosd(self.precone) + self.precurveTip*sind(self.precone)
        outputs['R']            = self.R
        outputs['diameter']     = self.R*2
        '''

    def compute_partials(self, inputs, J):

        J_sub = np.array([[cosd(self.precone), sind(self.precone),
            (-self.Rtip*sind(self.precone) + self.precurveTip*sind(self.precone))*np.pi/180.0]])

        J['R', 'Rtip'] = J_sub[0][0]
        J['R', 'precurveTip'] = J_sub[0][1]
        J['R', 'precone'] = J_sub[0][2]
        J['diameter', 'Rtip'] = 2.0*J_sub[0][0]
        J['diameter', 'precurveTip'] = 2.0*J_sub[0][1]
        J['diameter', 'precone'] = 2.0*J_sub[0][2]
        J['diameter', 'R'] = 2.0
        '''

        

class CCBladePower(ExplicitComponent):
    def initialize(self):
        self.options.declare('naero')
        self.options.declare('npower')

        self.options.declare('n_aoa_grid')
        self.options.declare('n_Re_grid')

        
    def setup(self):
        self.naero = naero = self.options['naero']
        npower = self.options['npower']
        n_aoa_grid = self.options['n_aoa_grid']
        n_Re_grid  = self.options['n_Re_grid']

        """blade element momentum code"""

        # inputs
        self.add_input('Uhub', val=np.zeros(npower), units='m/s', desc='hub height wind speed')
        self.add_input('Omega', val=np.zeros(npower), units='rpm', desc='rotor rotation speed')
        self.add_input('pitch', val=np.zeros(npower), units='deg', desc='blade pitch setting')

        # outputs
        self.add_output('T', val=np.zeros(npower), units='N', desc='rotor aerodynamic thrust')
        self.add_output('Q', val=np.zeros(npower), units='N*m', desc='rotor aerodynamic torque')
        self.add_output('P', val=np.zeros(npower), units='W', desc='rotor aerodynamic power')

        
        # (potential) variables
        self.add_input('r', val=np.zeros(naero), units='m', desc='radial locations where blade is defined (should be increasing and not go all the way to hub or tip)')
        self.add_input('chord', val=np.zeros(naero), units='m', desc='chord length at each section')
        self.add_input('theta', val=np.zeros(naero),  units='deg', desc='twist angle at each section (positive decreases angle of attack)')
        self.add_input('Rhub', val=0.0, units='m', desc='hub radius')
        self.add_input('Rtip', val=0.0, units='m', desc='tip radius')
        self.add_input('hub_height', val=0.0, units='m', desc='hub height')
        self.add_input('precone', val=0.0, desc='precone angle', units='deg')
        self.add_input('tilt', val=0.0, desc='shaft tilt', units='deg')
        self.add_input('yaw', val=0.0, desc='yaw error', units='deg')

        # TODO: I've not hooked up the gradients for these ones yet.
        self.add_input('precurve', val=np.zeros(naero), units='m', desc='precurve at each section')
        self.add_input('precurveTip', val=0.0, units='m', desc='precurve at tip')

        # parameters
        self.add_input('airfoils_cl', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='lift coefficients, spanwise')
        self.add_input('airfoils_cd', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='drag coefficients, spanwise')
        self.add_input('airfoils_cm', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='moment coefficients, spanwise')
        self.add_input('airfoils_aoa', val=np.zeros((n_aoa_grid)), units='deg', desc='angle of attack grid for polars')
        self.add_input('airfoils_Re', val=np.zeros((n_Re_grid)), desc='Reynolds numbers of polars')
        # self.add_discrete_input('airfoils', val=[0]*naero, desc='CCAirfoil instances')
        self.add_discrete_input('nBlades', val=0, desc='number of blades')
        self.add_input('rho', val=0.0, units='kg/m**3', desc='density of air')
        self.add_input('mu', val=0.0, units='kg/(m*s)', desc='dynamic viscosity of air')
        self.add_input('shearExp', val=0.0, desc='shear exponent')
        self.add_discrete_input('nSector', val=4, desc='number of sectors to divide rotor face into in computing thrust and power')
        self.add_discrete_input('tiploss', val=True, desc='include Prandtl tip loss model')
        self.add_discrete_input('hubloss', val=True, desc='include Prandtl hub loss model')
        self.add_discrete_input('wakerotation', val=True, desc='include effect of wake rotation (i.e., tangential induction factor is nonzero)')
        self.add_discrete_input('usecd', val=True, desc='use drag coefficient in computing induction factors')

        #self.declare_partials(['P', 'T', 'Q'],['precone', 'tilt', 'hub_height', 'Rhub', 'Rtip', 'yaw',
        #                                       'Uhub', 'Omega', 'pitch', 'r', 'chord', 'theta',
        #                                       'precurve', 'precurveTip'])

        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        self.r = inputs['r']
        self.chord = inputs['chord']
        self.theta = inputs['theta']
        self.Rhub = inputs['Rhub']
        self.Rtip = inputs['Rtip']
        self.hub_height = inputs['hub_height']
        self.precone = inputs['precone']
        self.tilt = inputs['tilt']
        self.yaw = inputs['yaw']
        self.precurve = inputs['precurve']
        self.precurveTip = inputs['precurveTip']
        # self.airfoils = discrete_inputs['airfoils']
        self.B = discrete_inputs['nBlades']
        self.rho = inputs['rho']
        self.mu = inputs['mu']
        self.shearExp = inputs['shearExp']
        self.nSector = discrete_inputs['nSector']
        self.tiploss = discrete_inputs['tiploss']
        self.hubloss = discrete_inputs['hubloss']
        self.wakerotation = discrete_inputs['wakerotation']
        self.usecd = discrete_inputs['usecd']
        self.Uhub = inputs['Uhub']
        self.Omega = inputs['Omega']
        self.pitch = inputs['pitch']

        af = [None]*self.naero
        for i in range(self.naero):
            af[i] = CCAirfoil(inputs['airfoils_aoa'], inputs['airfoils_Re'], inputs['airfoils_cl'][:,i,:], inputs['airfoils_cd'][:,i,:], inputs['airfoils_cm'][:,i,:])
        
        self.ccblade = CCBlade(self.r, self.chord, self.theta, af, self.Rhub, self.Rtip, self.B,
            self.rho, self.mu, self.precone, self.tilt, self.yaw, self.shearExp, self.hub_height,
            self.nSector, self.precurve, self.precurveTip, tiploss=self.tiploss, hubloss=self.hubloss,
            wakerotation=self.wakerotation, usecd=self.usecd, derivatives=True)

        # power, thrust, torque
        self.P, self.T, self.Q, self.M, self.dP, self.dT, self.dQ \
            = self.ccblade.evaluate(self.Uhub, self.Omega, self.pitch, coefficients=False)
        outputs['T'] = self.T
        outputs['Q'] = self.Q
        outputs['P'] = self.P
        '''

    def compute_partials(self, inputs, J, discrete_inputs):

        dP = self.dP
        dT = self.dT
        dQ = self.dQ
        
        J['P', 'precone']     = dP['dprecone']
        J['P', 'tilt']        = dP['dtilt']
        J['P', 'hub_height']  = dP['dhubHt']
        J['P', 'Rhub']        = dP['dRhub']
        J['P', 'Rtip']        = dP['dRtip']
        J['P', 'yaw']         = dP['dyaw']
        J['P', 'Uhub']        = dP['dUinf']
        J['P', 'Omega']       = dP['dOmega']
        J['P', 'pitch']       = dP['dpitch']
        J['P', 'r']           = dP['dr']
        J['P', 'chord']       = dP['dchord']
        J['P', 'theta']       = dP['dtheta']
        J['P', 'precurve']    = dP['dprecurve']
        J['P', 'precurveTip'] = dP['dprecurveTip']

        J['T', 'precone']     = dT['dprecone']
        J['T', 'tilt']        = dT['dtilt']
        J['T', 'hub_height']  = dT['dhubHt']
        J['T', 'Rhub']        = dT['dRhub']
        J['T', 'Rtip']        = dT['dRtip']
        J['T', 'yaw']         = dT['dyaw']
        J['T', 'Uhub']        = dT['dUinf']
        J['T', 'Omega']       = dT['dOmega']
        J['T', 'pitch']       = dT['dpitch']
        J['T', 'r']           = dT['dr']
        J['T', 'chord']       = dT['dchord']
        J['T', 'theta']       = dT['dtheta']
        J['T', 'precurve']    = dT['dprecurve']
        J['T', 'precurveTip'] = dT['dprecurveTip']

        J['Q', 'precone']     = dQ['dprecone']
        J['Q', 'tilt']        = dQ['dtilt']
        J['Q', 'hub_height']  = dQ['dhubHt']
        J['Q', 'Rhub']        = dQ['dRhub']
        J['Q', 'Rtip']        = dQ['dRtip']
        J['Q', 'yaw']         = dQ['dyaw']
        J['Q', 'Uhub']        = dQ['dUinf']
        J['Q', 'Omega']       = dQ['dOmega']
        J['Q', 'pitch']       = dQ['dpitch']
        J['Q', 'r']           = dQ['dr']
        J['Q', 'chord']       = dQ['dchord']
        J['Q', 'theta']       = dQ['dtheta']
        J['Q', 'precurve']    = dQ['dprecurve']
        J['Q', 'precurveTip'] = dQ['dprecurveTip']
        '''

        


    
class CCBladeLoads(ExplicitComponent):
    def initialize(self):
        self.options.declare('naero')
        self.options.declare('npower')

        self.options.declare('n_aoa_grid')
        self.options.declare('n_Re_grid')
        
    def setup(self):
        self.naero = naero = self.options['naero']
        npower     = self.options['npower']
        n_aoa_grid = self.options['n_aoa_grid']
        n_Re_grid  = self.options['n_Re_grid']
        """blade element momentum code"""

        # inputs
        self.add_input('V_load', val=0.0, units='m/s', desc='hub height wind speed')
        self.add_input('Omega_load', val=0.0, units='rpm', desc='rotor rotation speed')
        self.add_input('pitch_load', val=0.0, units='deg', desc='blade pitch setting')
        self.add_input('azimuth_load', val=0.0, units='deg', desc='blade azimuthal location')

        # outputs
        self.add_output('loads_r', val=np.zeros(naero), units='m', desc='radial positions along blade going toward tip')
        self.add_output('loads_Px', val=np.zeros(naero), units='N/m', desc='distributed loads in blade-aligned x-direction')
        self.add_output('loads_Py', val=np.zeros(naero), units='N/m', desc='distributed loads in blade-aligned y-direction')
        self.add_output('loads_Pz', val=np.zeros(naero), units='N/m', desc='distributed loads in blade-aligned z-direction')

        # corresponding setting for loads
        self.add_output('loads_V', val=0.0, units='m/s', desc='hub height wind speed')
        self.add_output('loads_Omega', val=0.0, units='rpm', desc='rotor rotation speed')
        self.add_output('loads_pitch', val=0.0, units='deg', desc='pitch angle')
        self.add_output('loads_azimuth', val=0.0, units='deg', desc='azimuthal angle')
        
        # (potential) variables
        self.add_input('r', val=np.zeros(naero), units='m', desc='radial locations where blade is defined (should be increasing and not go all the way to hub or tip)')
        self.add_input('chord', val=np.zeros(naero), units='m', desc='chord length at each section')
        self.add_input('theta', val=np.zeros(naero),  units='deg', desc='twist angle at each section (positive decreases angle of attack)')
        self.add_input('Rhub', val=0.0, units='m', desc='hub radius')
        self.add_input('Rtip', val=0.0, units='m', desc='tip radius')
        self.add_input('hub_height', val=0.0, units='m', desc='hub height')
        self.add_input('precone', val=0.0, desc='precone angle', units='deg')
        self.add_input('tilt', val=0.0, desc='shaft tilt', units='deg')
        self.add_input('yaw', val=0.0, desc='yaw error', units='deg')

        # TODO: I've not hooked up the gradients for these ones yet.
        self.add_input('precurve', val=np.zeros(naero), units='m', desc='precurve at each section')
        self.add_input('precurveTip', val=0.0, units='m', desc='precurve at tip')

        # parameters
        # self.add_discrete_input('airfoils', val=[0]*naero, desc='CCAirfoil instances')
        self.add_input('airfoils_cl', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='lift coefficients, spanwise')
        self.add_input('airfoils_cd', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='drag coefficients, spanwise')
        self.add_input('airfoils_cm', val=np.zeros((n_aoa_grid, naero, n_Re_grid)), desc='moment coefficients, spanwise')
        self.add_input('airfoils_aoa', val=np.zeros((n_aoa_grid)), units='deg', desc='angle of attack grid for polars')
        self.add_input('airfoils_Re', val=np.zeros((n_Re_grid)), desc='Reynolds numbers of polars')

        self.add_discrete_input('nBlades', val=0, desc='number of blades')
        self.add_input('rho', val=0.0, units='kg/m**3', desc='density of air')
        self.add_input('mu', val=0.0, units='kg/(m*s)', desc='dynamic viscosity of air')
        self.add_input('shearExp', val=0.0, desc='shear exponent')
        self.add_discrete_input('nSector', val=4, desc='number of sectors to divide rotor face into in computing thrust and power')
        self.add_discrete_input('tiploss', val=True, desc='include Prandtl tip loss model')
        self.add_discrete_input('hubloss', val=True, desc='include Prandtl hub loss model')
        self.add_discrete_input('wakerotation', val=True, desc='include effect of wake rotation (i.e., tangential induction factor is nonzero)')
        self.add_discrete_input('usecd', val=True, desc='use drag coefficient in computing induction factors')

        #self.declare_partials('loads_r', ['r', 'Rhub', 'Rtip'])
        #self.declare_partials(['loads_Px', 'loads_Py'],
        #                      ['r', 'chord', 'theta', 'Rhub', 'Rtip', 'hub_height', 'precone', 'tilt',
        #                       'yaw', 'V_load', 'Omega_load', 'pitch_load', 'azimuth_load', 'precurve'])
        #self.declare_partials('loads_V', 'V_load')
        #self.declare_partials('loads_Omega', 'Omega_load')
        #self.declare_partials('loads_pitch', 'pitch_load')
        #self.declare_partials('loads_azimuth', 'azimuth_load')

        
    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        self.r = inputs['r']
        self.chord = inputs['chord']
        self.theta = inputs['theta']
        self.Rhub = inputs['Rhub']
        self.Rtip = inputs['Rtip']
        self.hub_height = inputs['hub_height']
        self.precone = inputs['precone']
        self.tilt = inputs['tilt']
        self.yaw = inputs['yaw']
        self.precurve = inputs['precurve']
        self.precurveTip = inputs['precurveTip']
        # self.airfoils = discrete_inputs['airfoils']
        self.B = discrete_inputs['nBlades']
        self.rho = inputs['rho']
        self.mu = inputs['mu']
        self.shearExp = inputs['shearExp']
        self.nSector = discrete_inputs['nSector']
        self.tiploss = discrete_inputs['tiploss']
        self.hubloss = discrete_inputs['hubloss']
        self.wakerotation = discrete_inputs['wakerotation']
        self.usecd = discrete_inputs['usecd']
        self.V_load = inputs['V_load']
        self.Omega_load = inputs['Omega_load']
        self.pitch_load = inputs['pitch_load']
        self.azimuth_load = inputs['azimuth_load']


        if len(self.precurve) == 0:
            self.precurve = np.zeros_like(self.r)

        # airfoil files
        # n = len(self.airfoils)
        af = [None]*self.naero
        for i in range(self.naero):
            af[i] = CCAirfoil(inputs['airfoils_aoa'], inputs['airfoils_Re'], inputs['airfoils_cl'][:,i,:], inputs['airfoils_cd'][:,i,:], inputs['airfoils_cm'][:,i,:])
        # af = self.airfoils

        self.ccblade = CCBlade(self.r, self.chord, self.theta, af, self.Rhub, self.Rtip, self.B,
            self.rho, self.mu, self.precone, self.tilt, self.yaw, self.shearExp, self.hub_height,
            self.nSector, self.precurve, self.precurveTip, tiploss=self.tiploss, hubloss=self.hubloss,
            wakerotation=self.wakerotation, usecd=self.usecd, derivatives=True)

        # distributed loads
        Np, Tp, self.dNp, self.dTp \
            = self.ccblade.distributedAeroLoads(self.V_load, self.Omega_load, self.pitch_load, self.azimuth_load)
        
        
        # concatenate loads at root/tip
        outputs['loads_r'] = self.r

        # conform to blade-aligned coordinate system
        outputs['loads_Px'] = Np
        outputs['loads_Py'] = -Tp
        outputs['loads_Pz'] = 0*Np

        # return other outputs needed
        outputs['loads_V'] = self.V_load
        outputs['loads_Omega'] = self.Omega_load
        outputs['loads_pitch'] = self.pitch_load
        outputs['loads_azimuth'] = self.azimuth_load
        '''
    def compute_partials(self, inputs, J, discrete_inputs=None):

        dNp = self.dNp
        dTp = self.dTp
        n = len(self.r)

        dr_dr = np.eye(n)
        dr_dRhub = np.zeros(n)
        dr_dRtip = np.zeros(n)
        dr_dRhub[0] = 1.0
        dr_dRtip[-1] = 1.0

        dV = np.zeros(4*n+10)
        dV[3*n+6] = 1.0
        dOmega = np.zeros(4*n+10)
        dOmega[3*n+7] = 1.0
        dpitch = np.zeros(4*n+10)
        dpitch[3*n+8] = 1.0
        dazimuth = np.zeros(4*n+10)
        dazimuth[3*n+9] = 1.0

        zero = np.zeros(self.naero)
        J['loads_r',      'r']             = dr_dr
        J['loads_r',      'Rhub']          = dr_dRhub
        J['loads_r',      'Rtip']          = dr_dRtip
        J['loads_Px',     'r']             = dNp['dr']
        J['loads_Px',     'chord']         = dNp['dchord']
        J['loads_Px',     'theta']         = dNp['dtheta']
        J['loads_Px',     'Rhub']          = np.squeeze(dNp['dRhub'])
        J['loads_Px',     'Rtip']          = np.squeeze(dNp['dRtip'])
        J['loads_Px',     'hub_height']    = np.squeeze(dNp['dhubHt'])
        J['loads_Px',     'precone']       = np.squeeze(dNp['dprecone'])
        J['loads_Px',     'tilt']          = np.squeeze(dNp['dtilt'])
        J['loads_Px',     'yaw']           = np.squeeze(dNp['dyaw'])
        J['loads_Px',     'V_load']        = np.squeeze(dNp['dUinf'])
        J['loads_Px',     'Omega_load']    = np.squeeze(dNp['dOmega'])
        J['loads_Px',     'pitch_load']    = np.squeeze(dNp['dpitch'])
        J['loads_Px',     'azimuth_load']  = np.squeeze(dNp['dazimuth'])
        J['loads_Px',     'precurve']      = dNp['dprecurve']
        J['loads_Py',     'r']             = -dTp['dr']
        J['loads_Py',     'chord']         = -dTp['dchord']
        J['loads_Py',     'theta']         = -dTp['dtheta']
        J['loads_Py',     'Rhub']          = -np.squeeze(dTp['dRhub'])
        J['loads_Py',     'Rtip']          = -np.squeeze(dTp['dRtip'])
        J['loads_Py',     'hub_height']    = -np.squeeze(dTp['dhubHt'])
        J['loads_Py',     'precone']       = -np.squeeze(dTp['dprecone'])
        J['loads_Py',     'tilt']          = -np.squeeze(dTp['dtilt'])
        J['loads_Py',     'yaw']           = -np.squeeze(dTp['dyaw'])
        J['loads_Py',     'V_load']        = -np.squeeze(dTp['dUinf'])
        J['loads_Py',     'Omega_load']    = -np.squeeze(dTp['dOmega'])
        J['loads_Py',     'pitch_load']    = -np.squeeze(dTp['dpitch'])
        J['loads_Py',     'azimuth_load']  = -np.squeeze(dTp['dazimuth'])
        J['loads_Py',     'precurve']      = -dTp['dprecurve']
        J['loads_V',      'V_load']        = 1.0
        J['loads_Omega',  'Omega_load']    = 1.0
        J['loads_pitch',  'pitch_load']    = 1.0
        J['loads_azimuth', 'azimuth_load'] = 1.0
        '''

        




import numpy as np
from scipy.interpolate import griddata
from numpy.linalg import norm


def project_points(points,x,N_vect):
    """
    Rotate the surface and the points in order to align the vector N_vect in the z direction
    """

    N_vect = N_vect/norm(N_vect)
    rot = calculate_rotation_matrix(N_vect)
    points_rot = dotX(rot,points)
    x_rot = dotX(rot,x)
    points_rot2 = np.array([points_rot[:,0], points_rot[:,1],
                 griddata((x_rot[:,:,0].flatten(),x_rot[:,:,1].flatten()),x_rot[:,:,2].flatten(),
                      (points_rot[:,0],points_rot[:,1]), method = 'linear') ]).T 
    ### Rotate back
    inv_rot = np.linalg.inv(rot)
    points_final = dotX(inv_rot,points_rot2)
    return points_final

def normalize(v):

    return v / (np.dot(v,v)**2+1.e-16)

def curvature(points):

    if len(points.shape) < 2:
        return None
    if points.shape[1] == 1:
        return None

    if points.shape[1] == 2:
        d1 = np.diff(points.T)
        d2 = np.diff(d1)
        x1 = d1[0,1:]; y1 = d1[1,1:]
        x2 = d2[0,:];  y2 = d2[1,:]
        curv = (x1*y2-y1*x2)/(x1**2+y1**2)**(3./2.)

    elif points.shape[1] == 3:
        d1 = np.diff(points.T)
        d2 = np.diff(d1)
        x1 = d1[0,1:]; y1 = d1[1,1:]; z1 = d1[2,1:]
        x2 = d2[0,:]; y2 = d2[1,:]; z2 = d2[2,:]
        curv = ((z2*y1 - y2*z1)**2.0 + (x2*z1 - z2*x1)**2.0 + (y2*x1 - x2*y1)**2.0)**0.5/(x1**2.0 + y1**2.0 + z1**2.0+1.e-30)**(3.0/2.0)

    curvt = np.zeros(points.shape[0])
    try:
        curvt[1:-1] = curv
        curvt[0] = curv[0]
        curvt[-1] = curv[-1]
    except:
        pass
    return curvt

def calculate_angle(v1,v2):
    """
    Calculate the signed angle between the vector \e v1 and the vector \e v2

    \param    v1      <c>array(3)</c>   :       vector 1
    \param    v2      <c>array(3)</c>   :       vector 2
    \retval   angle   <c>float radian</c> :     the angle in the two vector plane
    """
    v1 = v1/norm(v1)
    v2 = v2/norm(v2)
    return np.arctan2(norm(np.cross(v1,v2)),np.dot(v1,v2))

def calculate_length(In):
    """
    Calculate the running length of a 3D line: (x,3) array 

    \param    In    the array of 3d points              <c> numpy.array((n,3)) </c>
    \retval   leng  a vector of the running distance    <c> numpy.array(n) </c>
    """
    if len(In.shape) == 1:
        d=np.diff(In)
        leng = np.cumsum(d)
        leng = np.insert(leng,0,0.)
        return leng

    ni, nj = In.shape
    if nj == 2:
        # segment lengths
        seglen  =np.zeros(In.shape[0]) 
        curvelen=np.zeros(In.shape[0]) 

        # calculate length and normals for each segment
        for i in range(1,In.shape[0]):
            seglen[i]=((In[i,0]-In[i-1,0])**2.0 \
                      +(In[i,1]-In[i-1,1])**2.0)**0.5
        # calculate accumulated curvelen
        leng=np.cumsum(seglen)

    elif nj ==3:
        seglen  =np.zeros(In.shape[0]) 
        curvelen=np.zeros(In.shape[0]) 
        for i in range(1,In.shape[0]):
            seglen[i]=((In[i,0]-In[i-1,0])**2.0 \
                      +(In[i,1]-In[i-1,1])**2.0 \
                      +(In[i,2]-In[i-1,2])**2.0)**0.5
        leng=np.cumsum(seglen)

    return leng

def calculate_rotation_matrix(vect):
    """
    Transpose (P1) and project the normal vector (vect) to the Z direction.

    \param    vect        vector to rotate                <c>numpy.array(3)</c>
    \retval   array       the rotation matrix             <c>numpy.array((3,3))</c>
    """
    return rotation_matrix_global(vect,np.array([0,0,1]))

def rotation_matrix_global(vect,direction):
    """
    Transpose (P1) and project the normal vector (vect) to the direction.

    \param    vect        vector to rotate               <c>numpy.array(3)</c>
    \param    direction   direction to rotate towards    <c>numpy.array(3)</c>
    \retval   array       the rotation matrix            <c>numpy.array((3,3))</c>
    """
    ### Normalize vect
    vect_norm = vect/norm(vect)
    ### Calculate the vector normal to the normal vector and the direction
    w = np.cross(vect_norm, direction)
    if norm(w) == 0:
        ### The two vectors are coplanar, exit with an identity matrix
        if np.dot(vect_norm, direction) > 0:
            return np.array([[1,0,0],[0,1,0],[0,0,1]])
        else:
            return -np.array([[1,0,0],[0,1,0],[0,0,1]])

    w_norm = w/norm(w)
    #print 'w_norm',w_norm
    ### The angle is found by taking the arccos of vect_norm[2]
    #q = math.acos(vect_norm[2])
    q = calculate_angle(vect_norm,direction)
    #print 'q',q
    ### calculate the rotation matrix of the vector w_norm and angle q
    rot = RotMat(w_norm,q)
    return rot

def dotX(rot, x, trans_vect=np.array([0.0,0.0,0.0])):
    """ 
    Transpose and Multiply the x array by a rotational matrix 
    """
    if isinstance(x,list):
        x_tmp = np.array([x[0].flatten(), x[1].flatten(), x[2].flatten()]).T 
        x_rot_tmp = np.zeros(x_tmp.shape)
        for i in range(x_tmp.shape[0]):
            x_rot_tmp[i,:] = dot(rot, x_tmp[i,:] - trans_vect)

        x_rot = []
        for iX in range(3):
            x_rot.append(x_rot_tmp[:,iX].reshape(x[0].shape))
    elif isinstance(x,np.ndarray):
        x_rot = np.zeros(x.shape)
        if len(x.shape)==1:
            x_rot = np.dot(rot, x - trans_vect)
        elif len(x.shape)==2:
            for i in range(x.shape[0]):
                x_rot[i,:] = np.dot(rot, x[i,:] - trans_vect)
        elif len(x.shape)==3:
            for i in range(x.shape[0]):
                for j in range(x.shape[1]):
                    x_rot[i,j,:] = np.dot(rot, x[i,j,:] - trans_vect)

    return x_rot

# rotation matrix function for an x-rotation
RotX = lambda a: np.array([[1,         0,        0     ], 
                        [0,       np.cos(a),   -np.sin(a)  ],
                        [0,       np.sin(a),   np.cos(a)  ]])
# rotation matrix function for a y-rotation
RotY = lambda a: np.array([[np.cos(a),    0,      np.sin(a)  ], 
                        [0,         1,        0     ], 
                        [-np.sin(a),   0,      np.cos(a)  ]])
# rotation matrix function for a z-rotation
RotZ = lambda a: np.array([[np.cos(a),  -np.sin(a),    0     ], 
                        [np.sin(a),   np.cos(a),    0     ], 
                        [0,         0,        1     ]])

def RotMat(u, theta):
    """ 
    Calculate the rotation matrix from a unit vector \e u and an angle \e theta
    
    http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    \f[
        R =\left[ \begin{array}{ccc}
        \cos\theta +u_x^2 \left(1-\cos \theta\right) & 
        u_x u_y \left(1-\cos \theta\right) - u_z \sin \theta & 
        u_x u_z \left(1-\cos \theta\right) + u_y \sin \theta \\ 
        u_y u_x \left(1-\cos \theta\right) + u_z \sin \theta & 
        \cos \theta + u_y^2\left(1-\cos \theta\right) & 
        u_y u_z \left(1-\cos \theta\right) - u_x \sin \theta \\
        u_z u_x \left(1-\cos \theta\right) - u_y \sin \theta &
        u_z u_y \left(1-\cos \theta\right) + u_x \sin \theta &
        \cos \theta + u_z^2\left(1-\cos \theta\right)
        \end{array} \right]
    \f]

    \param    u       <c> list/tuple/array(3) </c>       vector of the direction to rotate from
    \param    theta    <c> int radian </c>               angle to rotate with

    \retval   array   <c> array(3,3) </c>                rotation matrix
    """
    from numpy import cos, sin, array
    rot =  array([[cos(theta) + u[0]**2*(1-cos(theta)), u[0]*u[1]*(1-cos(theta)) - u[2]*sin(theta), u[0]*u[2]*(1-cos(theta)) + u[1]*sin(theta)], \
                  [u[1]*u[0]*(1-cos(theta)) + u[2]*sin(theta), cos(theta) + u[1]**2*(1-cos(theta)), u[1]*u[2]*(1-cos(theta)) - u[0]*sin(theta)], \
                  [u[2]*u[0]*(1-cos(theta)) - u[1]*sin(theta), u[2]*u[1]*(1-cos(theta)) + u[0]*sin(theta), cos(theta) + u[2]**2*(1-cos(theta))]])
    return rot

def dotXC(rot, x, center):
    """ 
    Transpose and Multiply the x array by a rotational matrix around a center
    """
    from numpy import zeros,array,dot
    if isinstance(x,list):
        x_tmp = array([x[0].flatten(), x[1].flatten(), x[2].flatten()]).T 
        x_rot_tmp = zeros(x_tmp.shape)
        for i in range(x_tmp.shape[0]):
            x_rot_tmp[i,:] = dot(rot, x_tmp[i,:] - center) + center

        x_rot = []
        for iX in range(3):
            x_rot.append(x_rot_tmp[:,iX].reshape(x[0].shape))
    elif isinstance(x,np.ndarray):
        x_rot = zeros(x.shape)        
        if len(x.shape)==1:
            x_rot[:] = dot(rot, x - center) + center

        if len(x.shape)==2:
            for i in range(x.shape[0]):
                x_rot[i,:] = dot(rot, x[i,:] - center) + center
        elif len(x.shape)==3:
            for i in range(x.shape[0]):
                for j in range(x.shape[1]):
                    x_rot[i,j,:] = dot(rot, x[i,j,:] - center) + center
    return x_rot



"""
drivese_omdao.py

Created by Yi Guo, Taylor Parsons and Ryan King 2014.
Copyright (c) NREL. All rights reserved.

Functions nacelle_example_5MW_baseline_[34]pt() did not define blade_mass
  We've added prob['blade_mass'] = 17740.0 (copied from hubse_omdao.py)
  GNS 2019 05 13
  
GNS 2019 06 05: nacelle_example_*() now return prob
  
Classes with declarations like
  class ObjName_OM(ExplicitComponent)
are OpenMDAO wrappers for pure-python objects that define the parts of a wind turbine drivetrain.
These objects are defined in drivese_components.py (which contains NO OpenMDAO code).
"""

import numpy as np
import sys

import wisdem.drivetrainse.drivese_components as dc
from wisdem.drivetrainse.hubse_omdao import HubSE, Hub_CM_Adder_OM
from wisdem.commonse.csystem import DirectionVector
from wisdem.commonse import gravity
from wisdem.commonse.utilities import hstack, vstack
from openmdao.api import Group, ExplicitComponent, IndepVarComp, Problem, view_connections

#-------------------------------------------------------------------------
# Components

#--------------------------------------------
class ForceMomentDemux(ExplicitComponent):
    def setup(self):
        self.add_input('Fxyz', val=np.zeros(3), units='N', desc='Force components applied at the hub center')
        self.add_input('Mxyz', val=np.zeros(3), units='N*m', desc='Moment components generated by the rotor')

        self.add_output('rotor_bending_moment_x', val=0.0, units='N*m', desc='The bending moment about the x axis')
        self.add_output('rotor_bending_moment_y', val=0.0, units='N*m', desc='The bending moment about the y axis')
        self.add_output('rotor_bending_moment_z', val=0.0, units='N*m', desc='The bending moment about the z axis')
        self.add_output('rotor_thrust',           val=0.0, units='N',   desc='The force along the x axis applied at hub center')
        self.add_output('rotor_force_y',          val=0.0, units='N',   desc='The force along the y axis applied at hub center')
        self.add_output('rotor_force_z',          val=0.0, units='N',   desc='The force along the z axis applied at hub center')

    def compute(self, inputs, outputs):
        outputs['rotor_thrust'], outputs['rotor_force_y'], outputs['rotor_force_z'] = inputs['Fxyz'][0], inputs['Fxyz'][1], inputs['Fxyz'][2]
        outputs['rotor_bending_moment_x'], outputs['rotor_bending_moment_y'], outputs['rotor_bending_moment_z'] = inputs['Mxyz'][0], inputs['Mxyz'][1], inputs['Mxyz'][2]

        
#-------------------------------------------------------------------------
class LowSpeedShaft4pt_OM(ExplicitComponent):
    ''' LowSpeedShaft class
          The LowSpeedShaft class is used to represent the low speed shaft component of a wind turbine drivetrain. 
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('rotor_bending_moment_x', val=0.0, units='N*m', desc='The bending moment about the x axis')
        self.add_input('rotor_bending_moment_y', val=0.0, units='N*m', desc='The bending moment about the y axis')
        self.add_input('rotor_bending_moment_z', val=0.0, units='N*m', desc='The bending moment about the z axis')
        self.add_input('rotor_thrust',           val=0.0, units='N',   desc='The force along the x axis applied at hub center')
        self.add_input('rotor_force_y',          val=0.0, units='N',   desc='The force along the y axis applied at hub center')
        self.add_input('rotor_force_z',          val=0.0, units='N',   desc='The force along the z axis applied at hub center')
        self.add_input('rotor_mass',             val=0.0, units='kg',  desc='rotor mass')
        self.add_input('rotor_diameter',         val=0.0, units='m',   desc='rotor diameter')
        self.add_input('machine_rating',         val=0.0, units='kW',  desc='machine_rating machine rating of the turbine')
        self.add_input('gearbox_mass',           val=0.0, units='kg',  desc='Gearbox mass')
        self.add_input('carrier_mass',           val=0.0, units='kg',  desc='Carrier mass')
        self.add_input('overhang',               val=0.0, units='m',   desc='Overhang distance')
        self.add_input('distance_hub2mb',        val=0.0, units='m',   desc='distance between hub center and upwind main bearing')
        self.add_input('drivetrain_efficiency',  val=0.0,              desc='overall drivettrain efficiency')

        # parameters
        self.add_input('shrink_disc_mass', val=0.0,         units='kg',  desc='Mass of the shrink disc')
        self.add_input('gearbox_cm',       val=np.zeros(3), units='m',   desc='center of mass of gearbox')
        self.add_input('gearbox_length',   val=0.0,         units='m',   desc='gearbox length')
        self.add_input('flange_length',    val=0.0,         units='m',   desc='flange length')
        self.add_input('shaft_angle',      val=0.0,         units='rad', desc='Angle of the LSS inclination with respect to the horizontal')
        self.add_input('shaft_ratio',      val=0.0,                      desc='Ratio of inner diameter to outer diameter.  Leave zero for solid LSS')
      
        self.add_input('hub_flange_thickness', val=0.0, desc='Shell thickness for spherical hub')

        self.add_discrete_input('mb1Type', val='CARB', desc='main bearing #1 type- valid options are CRB/SRB/RB/CARB/TRB1/TRB2')
        self.add_discrete_input('mb2Type', val='SRB', desc='main bearing #2 type- valid options are CRB/SRB/RB/CARB/TRB1/TRB2')
        self.add_discrete_input('IEC_Class', val='B', desc='IEC turbulence class (A/B/C)')
        
        # outputs
        self.add_output('lss_design_torque',       val=0.0,         units='N*m', desc='lss design torque')
        self.add_output('lss_design_bending_load', val=0.0,         units='N',   desc='lss design bending load')
        self.add_output('lss_length',              val=0.0,         units='m',   desc='lss length')
        self.add_output('lss_diameter1',           val=0.0,         units='m',   desc='lss outer diameter at main bearing')
        self.add_output('lss_diameter2',           val=0.0,         units='m',   desc='lss outer diameter at second bearing')
        self.add_output('lss_mass',                val=0.0,         units='kg',  desc='overall component mass')
        self.add_output('lss_cm',                  val=np.zeros(3), units='m',   desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('lss_I',                   val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('lss_mb1_facewidth',       val=0.0,         units='m',   desc='facewidth of upwind main bearing')
        self.add_output('lss_mb2_facewidth',       val=0.0,         units='m',   desc='facewidth of main bearing')
        self.add_output('lss_mb1_mass',            val=0.0,         units='kg',  desc='main bearing mass')
        self.add_output('lss_mb2_mass',            val=0.0,         units='kg',  desc='second bearing mass')
        self.add_output('lss_mb1_cm',              val=np.zeros(3), units='m',   desc='main bearing 1 center of mass')
        self.add_output('lss_mb2_cm',              val=np.zeros(3), units='m',   desc='main bearing 2 center of mass')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        lss4pt = dc.LowSpeedShaft4pt(discrete_inputs['mb1Type'], discrete_inputs['mb2Type'], discrete_inputs['IEC_Class'], debug=self.options['debug'])

        (outputs['lss_design_torque'], outputs['lss_design_bending_load'], outputs['lss_length'], outputs['lss_diameter1'], outputs['lss_diameter2'], outputs['lss_mass'], outputs['lss_cm'], outputs['lss_I'], \
         outputs['lss_mb1_facewidth'], outputs['lss_mb2_facewidth'], outputs['lss_mb1_mass'], outputs['lss_mb2_mass'], outputs['lss_mb1_cm'], outputs['lss_mb2_cm']) \
                = lss4pt.compute(inputs['rotor_diameter'], inputs['rotor_mass'], inputs['rotor_thrust'], inputs['rotor_force_y'], inputs['rotor_force_z'], \
                                    inputs['rotor_bending_moment_x'], inputs['rotor_bending_moment_y'], inputs['rotor_bending_moment_z'], \
                                    inputs['overhang'], inputs['machine_rating'], inputs['drivetrain_efficiency'], \
                                    inputs['gearbox_mass'], inputs['carrier_mass'], inputs['gearbox_cm'], inputs['gearbox_length'], \
                                    inputs['shrink_disc_mass'], inputs['flange_length'], inputs['distance_hub2mb'], inputs['shaft_angle'], inputs['shaft_ratio'], \
                                    inputs['hub_flange_thickness'])

        

#-------------------------------------------------------------------------


class LowSpeedShaft3pt_OM(ExplicitComponent):
    ''' LowSpeedShaft class
          The LowSpeedShaft class is used to represent the low speed shaft component of a wind turbine drivetrain. 
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('rotor_bending_moment_x', val=0.0, units='N*m', desc='The bending moment about the x axis')
        self.add_input('rotor_bending_moment_y', val=0.0, units='N*m', desc='The bending moment about the y axis')
        self.add_input('rotor_bending_moment_z', val=0.0, units='N*m', desc='The bending moment about the z axis')
        self.add_input('rotor_thrust',           val=0.0, units='N',   desc='The force along the x axis applied at hub center')
        self.add_input('rotor_force_y',          val=0.0, units='N',   desc='The force along the y axis applied at hub center')
        self.add_input('rotor_force_z',          val=0.0, units='N',   desc='The force along the z axis applied at hub center')
        self.add_input('rotor_mass',             val=0.0, units='kg',  desc='rotor mass')
        self.add_input('rotor_diameter',         val=0.0, units='m',   desc='rotor diameter')
        self.add_input('machine_rating',         val=0.0, units='kW',  desc='machine_rating machine rating of the turbine')
        self.add_input('gearbox_mass',           val=0.0, units='kg',  desc='Gearbox mass')
        self.add_input('carrier_mass',           val=0.0, units='kg',  desc='Carrier mass')
        self.add_input('overhang',               val=0.0, units='m',   desc='Overhang distance')
        self.add_input('distance_hub2mb',        val=0.0, units='m',   desc='distance between hub center and upwind main bearing')
        self.add_input('drivetrain_efficiency',  val=0.0,              desc='overall drivettrain efficiency')

        # parameters
        self.add_input('shrink_disc_mass', val=0.0,         units='kg',  desc='Mass of the shrink disc')
        self.add_input('gearbox_cm',       val=np.zeros(3), units='m',   desc='center of mass of gearbox')
        self.add_input('gearbox_length',   val=0.0,         units='m',   desc='gearbox length')
        self.add_input('flange_length',    val=0.0,         units='m',   desc='flange length')
        self.add_input('shaft_angle',      val=0.0,         units='rad', desc='Angle of the LSS inclination with respect to the horizontal')
        self.add_input('shaft_ratio',      val=0.0,                      desc='Ratio of inner diameter to outer diameter.  Leave zero for solid LSS')
        
        self.add_input('hub_flange_thickness', val=0.0, desc='Shell thickness for spherical hub')

        self.add_discrete_input('mb1Type', val='CARB', desc='main bearing #1 type- valid options are CRB/SRB/RB/CARB/TRB1/TRB2')
        self.add_discrete_input('IEC_Class', val='B', desc='IEC turbulence class (A/B/C)')
        
        # outputs
        self.add_output('lss_design_torque',       val=0.0,         units='N*m', desc='lss design torque')
        self.add_output('lss_design_bending_load', val=0.0,         units='N',   desc='lss design bending load')
        self.add_output('lss_length',              val=0.0,         units='m',   desc='lss length')
        self.add_output('lss_diameter1',           val=0.0,         units='m',   desc='lss outer diameter at main bearing')
        self.add_output('lss_diameter2',           val=0.0,         units='m',   desc='lss outer diameter at second bearing')
        self.add_output('lss_mass',                val=0.0,         units='kg',  desc='overall component mass')
        self.add_output('lss_cm',                  val=np.zeros(3),              desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('lss_I',                   val=np.zeros(3),              desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('lss_mb1_facewidth',       val=0.0,         units='m',   desc='facewidth of upwind main bearing')
        self.add_output('lss_mb2_facewidth',       val=0.0,         units='m',   desc='facewidth of main bearing')
        self.add_output('lss_mb1_mass',            val=0.0,         units='kg',  desc='main bearing mass')
        self.add_output('lss_mb2_mass',            val=0.0,         units='kg',  desc='second bearing mass')
        self.add_output('lss_mb1_cm',              val=np.zeros(3), units='m',   desc='main bearing 1 center of mass')
        self.add_output('lss_mb2_cm',              val=np.zeros(3), units='m',   desc='main bearing 2 center of mass')


    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        lss3pt = dc.LowSpeedShaft3pt(discrete_inputs['mb1Type'], discrete_inputs['IEC_Class'], debug=self.options['debug'])
        
        (outputs['lss_design_torque'], outputs['lss_design_bending_load'], outputs['lss_length'], outputs['lss_diameter1'], outputs['lss_diameter2'], outputs['lss_mass'], outputs['lss_cm'], outputs['lss_I'], \
         outputs['lss_mb1_facewidth'], outputs['lss_mb2_facewidth'], outputs['lss_mb1_mass'], outputs['lss_mb2_mass'], outputs['lss_mb1_cm'], outputs['lss_mb2_cm']) \
                = lss3pt.compute(inputs['rotor_diameter'], inputs['rotor_mass'], inputs['rotor_thrust'], inputs['rotor_force_y'], inputs['rotor_force_z'], \
                                    inputs['rotor_bending_moment_x'], inputs['rotor_bending_moment_y'], inputs['rotor_bending_moment_z'], \
                                    inputs['overhang'], inputs['machine_rating'], inputs['drivetrain_efficiency'], \
                                    inputs['gearbox_mass'], inputs['carrier_mass'], inputs['gearbox_cm'], inputs['gearbox_length'], \
                                    inputs['shrink_disc_mass'], inputs['flange_length'], inputs['distance_hub2mb'], inputs['shaft_angle'], inputs['shaft_ratio'],
                                    inputs['hub_flange_thickness'])       

        

#-------------------------------------------------------------------------

class MainBearing_OM(ExplicitComponent):
    ''' MainBearings class
          The MainBearings class is used to represent the main bearing components of a wind turbine drivetrain. It contains two subcomponents (main bearing and second bearing) which also inherit from the SubComponent class.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''

    def initialize(self):
        self.options.declare('bearing_position', default='main')
        
    def setup(self):
        # variables
        self.add_input('bearing_mass', val=0.0, units='kg', desc='bearing mass from LSS model')
        self.add_input('lss_diameter', val=0.0, units='m', desc='lss outer diameter at main bearing')
        self.add_input('lss_design_torque', val=0.0, units='N*m', desc='lss design torque')
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('lss_mb_cm', val=np.array([0., 0., 0.]), units='m', desc='x,y,z location from shaft model')

        # returns
        self.add_output('mb_mass', val=0.0, units='kg', desc='overall component mass')
        self.add_output('mb_cm',   val=np.zeros(3), units='m', desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('mb_I',    val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')

        
    def compute(self, inputs, outputs):

        mb = dc.MainBearing(self.options['bearing_position'])
        
        (outputs['mb_mass'], outputs['mb_cm'], outputs['mb_I']) \
            = mb.compute(inputs['bearing_mass'], inputs['lss_diameter'], inputs['lss_design_torque'], inputs['rotor_diameter'], inputs['lss_mb_cm'])

        

#-------------------------------------------------------------------------

class Gearbox_OM(ExplicitComponent):
    ''' Gearbox class
          The Gearbox class is used to represent the gearbox component of a wind turbine drivetrain.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''
    def initialize(self):
        self.options.declare('debug', default=False)
        
    def setup(self):
        # variables
        self.add_input('gear_ratio', val=0.0, desc='overall gearbox speedup ratio')
        self.add_discrete_input('planet_numbers', val=np.array([0, 0, 0,]), desc='number of planets in each stage')
        self.add_input('rotor_rpm', val=0.0, units='rpm', desc='rotor rpm at rated power')
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('rotor_torque', val=0.0, units='N*m', desc='rotor torque at rated power')
        self.add_input('gearbox_input_xcm', val=0.00, units='m', desc='gearbox position along x-axis')

        self.add_discrete_input('gear_configuration', val='eep', desc='string that represents the configuration of the gearbox (stage number and types)')
        self.add_discrete_input('shaft_factor', val='normal', desc='normal or short shaft length')
        
        # outputs
        self.add_output('stage_masses', val=np.zeros(3), units='kg', desc='individual gearbox stage gearbox_masses')
        self.add_output('gearbox_mass', val=0.0, units='kg', desc='overall component gearbox_mass')
        self.add_output('gearbox_cm', val=np.zeros(3), units='m', desc='center of gearbox_mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('gearbox_I', val=np.zeros(3), units='kg*m**2', desc=' moments of gearbox_Inertia for the component [gearbox_Ixx, gearbox_Iyy, gearbox_Izz] around its center of gearbox_mass')
        self.add_output('gearbox_length', val=0.0, units='m', desc='gearbox length')
        self.add_output('gearbox_height', val=0.0, units='m', desc='gearbox height')
        self.add_output('gearbox_diameter', val=0.0, units='m', desc='gearbox diameter')


    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        mygearbox = dc.Gearbox(discrete_inputs['gear_configuration'], discrete_inputs['shaft_factor'], debug=self.options['debug'])
        
        (outputs['stage_masses'], outputs['gearbox_mass'], outputs['gearbox_cm'], outputs['gearbox_I'], outputs['gearbox_length'], outputs['gearbox_height'], outputs['gearbox_diameter']) \
            = mygearbox.compute(inputs['gear_ratio'], discrete_inputs['planet_numbers'], inputs['rotor_rpm'], inputs['rotor_diameter'], inputs['rotor_torque'], inputs['gearbox_input_xcm'])

        

#-------------------------------------------------------------------

class HighSpeedSide_OM(ExplicitComponent):
    '''
    HighSpeedShaft class
          The HighSpeedShaft class is used to represent the high speed shaft and mechanical brake components of a wind turbine drivetrain.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''

    def setup(self):

        # variables
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('rotor_torque', val=0.0, units='N*m', desc='rotor torque at rated power')
        self.add_input('gear_ratio', val=0.0, desc='overall gearbox ratio')
        self.add_input('lss_diameter', val=0.0, units='m', desc='low speed shaft outer diameter')
        self.add_input('gearbox_length', val=0.0, units='m', desc='gearbox length')
        self.add_input('gearbox_height', val=0.0, units='m', desc='gearbox height')
        self.add_input('gearbox_cm', val=np.zeros(3), units='m', desc='gearbox cm [x,y,z]')
        self.add_input('hss_input_length', val=0.0, units='m', desc='high speed shaft length determined by user. Default 0.5m')

        # returns
        self.add_output('hss_mass', val=0.0, units='kg', desc='overall component mass')
        self.add_output('hss_cm', val=np.zeros(3), units='m', desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('hss_I', val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('hss_length', val=0.0, units='m', desc='length of high speed shaft')

        self.hss = dc.HighSpeedSide()

    def compute(self, inputs, outputs):

        (outputs['hss_mass'], outputs['hss_cm'], outputs['hss_I'], outputs['hss_length']) \
            = self.hss.compute(inputs['rotor_diameter'], inputs['rotor_torque'], inputs['gear_ratio'], inputs['lss_diameter'], inputs['gearbox_length'], inputs['gearbox_height'], inputs['gearbox_cm'], inputs['hss_input_length'])

        

#----------------------------------------------------------------------------------------------

class Generator_OM(ExplicitComponent):
    '''Generator class
          The Generator class is used to represent the generator of a wind turbine drivetrain.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''

        
    def setup(self):
        # variables
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('machine_rating', val=0.0, units='kW', desc='machine rating of generator')
        self.add_input('gear_ratio', val=0.0, desc='overall gearbox ratio')
        self.add_input('hss_length', val=0.0, units='m', desc='length of high speed shaft and brake')
        self.add_input('hss_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='cm of high speed shaft and brake')
        self.add_input('rotor_rpm', val=0.0, units='rpm', desc='Speed of rotor at rated power')
        
        self.add_discrete_input('drivetrain_design', val='geared', desc='geared or single_stage or multi_drive or pm_direct_drive')

        #returns
        self.add_output('generator_mass', val=0.0, units='kg', desc='overall component mass')
        self.add_output('generator_cm', val=np.zeros(3), units='m', desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('generator_I', val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        mygen = dc.Generator(discrete_inputs['drivetrain_design'])
        
        (outputs['generator_mass'], outputs['generator_cm'], outputs['generator_I']) \
            = mygen.compute(inputs['rotor_diameter'], inputs['machine_rating'], inputs['gear_ratio'], inputs['hss_length'], inputs['hss_cm'], inputs['rotor_rpm'])

        

        
        
#-------------------------------------------------------------------------------

class Transformer_OM(ExplicitComponent):
    ''' Transformer class
            The transformer class is used to represent the transformer of a wind turbine drivetrain.
            It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
            It contains an update method to determine the mass, mass properties, and dimensions of the component if it is in fact uptower'''

        
    def setup(self):

        # inputs
        self.add_input('machine_rating', val=0.0, units='kW', desc='machine rating of the turbine')
        self.add_input('tower_top_diameter', val=0.0, units='m', desc='tower top diameter for comparision of nacelle CM')
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter of turbine')
        self.add_input('rotor_mass', val=0.0, units='kg', desc='rotor mass')
        self.add_input('lss_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mb1_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mb2_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('gearbox_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('hss_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('generator_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('lss_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('mb1_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('mb2_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('gearbox_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('hss_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('generator_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('overhang', val=0.0, units='m', desc='nacelle overhang')

        self.add_discrete_input('uptower_transformer', val=True)
        
        # outputs
        self.add_output('transformer_mass', val=0.0, units='kg', desc='overall component mass')
        self.add_output('transformer_cm', val=np.zeros(3), units='m', desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('transformer_I', val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')    


    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        rnaadder = dc.RNASystemAdder()
        rna_mass, rna_cm = rnaadder.compute(inputs['lss_mass'], inputs['mb1_mass'], inputs['mb2_mass'], inputs['gearbox_mass'],
                                            inputs['hss_mass'], inputs['generator_mass'], inputs['lss_cm'], inputs['mb1_cm'],
                                            inputs['mb2_cm'], inputs['gearbox_cm'], inputs['hss_cm'], inputs['generator_cm'],
                                            inputs['overhang'], inputs['rotor_mass'], inputs['machine_rating'])

        transformer = dc.Transformer(discrete_inputs['uptower_transformer'])
        (outputs['transformer_mass'], outputs['transformer_cm'], outputs['transformer_I']) = transformer.compute(inputs['machine_rating'], inputs['tower_top_diameter'],
                                                                                                                 inputs['rotor_mass'], inputs['generator_cm'],
                                                                                                                 inputs['rotor_diameter'], rna_mass, rna_cm)

        

#-------------------------------------------------------------------------

class Bedplate_OM(ExplicitComponent):
    ''' Bedplate class
          The Bedplate class is used to represent the bedplate of a wind turbine drivetrain.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''
    def initialize(self):
        self.options.declare('debug', default=False)

    def setup(self):
        # variables
        self.add_input('gearbox_length', val=0.0, units='m', desc='gearbox length')
        self.add_input('gearbox_location', val=0.0, units='m', desc='gearbox CM location')
        self.add_input('gearbox_mass', val=0.0, units='kg', desc='gearbox mass')
        self.add_input('hss_location', val=0.0, units='m', desc='HSS CM location')
        self.add_input('hss_mass', val=0.0, units='kg', desc='HSS mass')
        self.add_input('generator_location', val=0.0, units='m', desc='generator CM location')
        self.add_input('generator_mass', val=0.0, units='kg', desc='generator mass')
        self.add_input('lss_location', val=0.0, units='m', desc='LSS CM location')
        self.add_input('lss_mass', val=0.0, units='kg', desc='LSS mass')
        self.add_input('lss_length', val=0.0, units='m', desc='LSS length')
        self.add_input('lss_mb1_facewidth', val=0.0, units='m', desc='Upwind main bearing facewidth')
        self.add_input('mb1_cm', val=np.zeros(3), units='m', desc='Upwind main bearing CM location')
        self.add_input('mb1_mass', val=0.0, units='kg', desc='Upwind main bearing mass')
        self.add_input('mb2_cm', val=np.zeros(3), units='m', desc='Downwind main bearing CM location')
        self.add_input('mb2_mass', val=0.0, units='kg', desc='Downwind main bearing mass')
        self.add_input('transformer_mass', val=0.0, units='kg', desc='Transformer mass')
        self.add_input('transformer_cm', val=np.zeros(3), units='m', desc='transformer CM location')
        self.add_input('tower_top_diameter', val=0.0, units='m', desc='diameter of the top tower section at the yaw gear')
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('machine_rating', val=0.0, units='kW', desc='machine_rating machine rating of the turbine')
        self.add_input('rotor_mass', val=0.0, units='kg', desc='rotor mass')
        self.add_input('rotor_bending_moment_y', val=0.0, units='N*m', desc='The bending moment about the y axis')
        self.add_input('rotor_force_z', val=0.0, units='N', desc='The force along the z axis applied at hub center')
        self.add_input('flange_length', val=0.0, units='m', desc='flange length')
        self.add_input('distance_hub2mb', val=0.0, units='m', desc='length between rotor center and upwind main bearing')

        self.add_discrete_input('uptower_transformer', val=True)
        
        # outputs
        self.add_output('bedplate_mass', val=0.0, units='kg', desc='overall component bedplate_mass')
        self.add_output('bedplate_cm', val=np.zeros(3), units='m', desc='center of bedplate_mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('bedplate_I', val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of bedplate_mass')
        self.add_output('bedplate_length', val=0.0, units='m', desc='length of bedplate')
        self.add_output('bedplate_height', val=0.0, units='m',  desc='max height of bedplate')
        self.add_output('bedplate_width', val=0.0, units='m', desc='width of bedplate')
        

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        bpl = dc.Bedplate(discrete_inputs['uptower_transformer'], debug=self.options['debug'])
        
        (outputs['bedplate_mass'], outputs['bedplate_cm'], outputs['bedplate_I'], outputs['bedplate_length'], outputs['bedplate_height'], outputs['bedplate_width']) \
            = bpl.compute(inputs['gearbox_length'], inputs['gearbox_location'], inputs['gearbox_mass'], inputs['hss_location'], inputs['hss_mass'], inputs['generator_location'], inputs['generator_mass'], \
                      inputs['lss_location'], inputs['lss_mass'], inputs['lss_length'], inputs['mb1_cm'], inputs['lss_mb1_facewidth'], inputs['mb1_mass'], inputs['mb2_cm'], inputs['mb2_mass'], \
                      inputs['transformer_mass'], inputs['transformer_cm'], \
                      inputs['tower_top_diameter'], inputs['rotor_diameter'], inputs['machine_rating'], inputs['rotor_mass'], inputs['rotor_bending_moment_y'], inputs['rotor_force_z'], \
                      inputs['flange_length'], inputs['distance_hub2mb'])

        

#-------------------------------------------------------------------------------

class AboveYawMassAdder_OM(ExplicitComponent):
    ''' AboveYawMassAdder_OM class
          The AboveYawMassAdder_OM class is used to represent the masses of all parts of a wind turbine drivetrain that
          are above the yaw system.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''
    def initialize(self):
        self.options.declare('debug', default=False)

    def setup(self):
        # variables
        self.add_input('machine_rating', val=0.0, units='kW', desc='machine rating')
        self.add_input('lss_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mb1_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mb2_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('gearbox_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('hss_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('generator_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('bedplate_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('bedplate_length', val=0.0, units='m', desc='component length')
        self.add_input('bedplate_width', val=0.0, units='m', desc='component width')
        self.add_input('transformer_mass', val=0.0, units='kg', desc='component mass')

        self.add_discrete_input('crane', val=True, desc='onboard crane present')
        
        # returns
        self.add_output('electrical_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('vs_electronics_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('hvac_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('controls_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('platforms_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('crane_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('mainframe_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('cover_mass', val=0.0, units='kg', desc='component mass')
        self.add_output('above_yaw_mass', val=0.0, units='kg', desc='total mass above yaw system')
        self.add_output('nacelle_length', val=0.0, units='m', desc='component length')
        self.add_output('nacelle_width', val=0.0, units='m', desc='component width')
        self.add_output('nacelle_height', val=0.0, units='m', desc='component height')
        

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        aboveyawmass = dc.AboveYawMassAdder(discrete_inputs['crane'])

        (outputs['electrical_mass'], outputs['vs_electronics_mass'], outputs['hvac_mass'], outputs['controls_mass'], 
         outputs['platforms_mass'], outputs['crane_mass'], outputs['mainframe_mass'], outputs['cover_mass'], 
         outputs['above_yaw_mass'], outputs['nacelle_length'], outputs['nacelle_width'], outputs['nacelle_height']) \
            = aboveyawmass.compute(inputs['machine_rating'], inputs['lss_mass'], inputs['mb1_mass'], inputs['mb2_mass'], 
                    inputs['gearbox_mass'], inputs['hss_mass'], inputs['generator_mass'], inputs['bedplate_mass'], 
                    inputs['bedplate_length'], inputs['bedplate_width'], inputs['transformer_mass'])
        
        if self.options['debug']:
            print('AYMA IN: {:.1f} kW BPl {:.1f} m BPw {:.1f} m'.format(
                  inputs['machine_rating'],inputs['bedplate_length'], inputs['bedplate_width']))
            print('AYMA IN  masses (kg): LSS {:.1f} MB1 {:.1f} MB2 {:.1f} GBOX {:.1f} HSS {:.1f} GEN {:.1f} BP {:.1f} TFRM {:.1f}'.format(
                  inputs['lss_mass'], inputs['mb1_mass'], inputs['mb2_mass'], inputs['gearbox_mass'],
                  inputs['hss_mass'], inputs['generator_mass'], inputs['bedplate_mass'], inputs['transformer_mass']))
            print('AYMA OUT masses (kg) : E {:.1f} VSE {:.1f} HVAC {:.1f} CNTL {:.1f} PTFM {:.1f} CRN {:.1f} MNFRM {:.1f} CVR {:.1f} AYM {:.1f}'.format( 
                  outputs['electrical_mass'], outputs['vs_electronics_mass'], outputs['hvac_mass'], outputs['controls_mass'],
                  outputs['platforms_mass'], outputs['crane_mass'], outputs['mainframe_mass'], outputs['cover_mass'],
                  outputs['above_yaw_mass']))
            print('AYMA OUT nacelle (m): L {:.2f} W {:.2f} H {:.2f}'.format( 
                 outputs['nacelle_length'], outputs['nacelle_width'], outputs['nacelle_height']))

        

#---------------------------------------------------------------------------------------------------------------

class YawSystem_OM(ExplicitComponent):
    ''' YawSystem class
          The YawSystem class is used to represent the yaw system of a wind turbine drivetrain.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''

    def setup(self):
        # variables
        self.add_input('rotor_diameter', val=0.0, units='m', desc='rotor diameter')
        self.add_input('rotor_thrust', val=0.0, units='N', desc='maximum rotor thrust')
        self.add_input('tower_top_diameter', val=0.0, units='m', desc='tower top diameter')
        self.add_input('above_yaw_mass', val=0.0, units='kg', desc='above yaw mass')
        self.add_input('bedplate_height', val=0.0, units='m', desc='bedplate height')

        self.add_discrete_input('yaw_motors_number', val=0, desc='default value - will be internally calculated')
        
        # outputs
        self.add_output('yaw_mass', val=0.0, units='kg', desc='overall component mass')
        self.add_output('yaw_cm', val=np.zeros(3), units='m', desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('yaw_I', val=np.zeros(3), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')    

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        myyaw = dc.YawSystem(discrete_inputs['yaw_motors_number'])

        (outputs['yaw_mass'], outputs['yaw_cm'], outputs['yaw_I']) \
            = myyaw.compute(inputs['rotor_diameter'], inputs['rotor_thrust'], inputs['tower_top_diameter'], inputs['above_yaw_mass'], inputs['bedplate_height'])

        

#--------------------------------------------
class NacelleSystemAdder_OM(ExplicitComponent): #added to drive to include transformer
    ''' NacelleSystem class
          The Nacelle class is used to represent the overall nacelle of a wind turbine.
          It contains the general properties for a wind turbine component as well as additional design load and dimensional attributes as listed below.
          It contains an update method to determine the mass, mass properties, and dimensions of the component.
    '''

    def setup(self):

        # variables
        self.add_input('above_yaw_mass', val=0.0, units='kg', desc='mass above yaw system')
        self.add_input('yaw_mass', val=0.0, units='kg', desc='mass of yaw system')
        self.add_input('lss_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mb1_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mb2_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('gearbox_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('hss_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('generator_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('bedplate_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('mainframe_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('lss_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('mb1_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('mb2_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('gearbox_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('hss_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('generator_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('bedplate_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('lss_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('mb1_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('mb2_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('gearbox_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('hss_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('generator_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('bedplate_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')
        self.add_input('transformer_mass', val=0.0, units='kg', desc='component mass')
        self.add_input('transformer_cm', val=np.array([0.0,0.0,0.0]), units='m', desc='component CM')
        self.add_input('transformer_I', val=np.array([0.0,0.0,0.0]), units='kg*m**2', desc='component I')

        # returns
        self.add_output('nacelle_mass', val=0.0, units='kg', desc='overall component mass')
        self.add_output('nacelle_cm', val=np.zeros(3), units='m', desc='center of mass of the component in [x,y,z] for an arbitrary coordinate system')
        self.add_output('nacelle_I', val=np.zeros(6), units='kg*m**2', desc=' moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')

        
    def compute(self, inputs, outputs):
        nacelleadder = dc.NacelleSystemAdder()

        (outputs['nacelle_mass'], outputs['nacelle_cm'], outputs['nacelle_I']) \
                    = nacelleadder.compute(inputs['above_yaw_mass'], inputs['yaw_mass'], inputs['lss_mass'], inputs['mb1_mass'], inputs['mb2_mass'], inputs['gearbox_mass'], \
                      inputs['hss_mass'], inputs['generator_mass'], inputs['bedplate_mass'], inputs['mainframe_mass'], \
                      inputs['lss_cm'], inputs['mb1_cm'], inputs['mb2_cm'], inputs['gearbox_cm'], inputs['hss_cm'], inputs['generator_cm'], inputs['bedplate_cm'], \
                      inputs['lss_I'], inputs['mb1_I'], inputs['mb2_I'], inputs['gearbox_I'], inputs['hss_I'], inputs['generator_I'], inputs['bedplate_I'], \
                      inputs['transformer_mass'], inputs['transformer_cm'], inputs['transformer_I'])


#--------------------------------------------
class RNAMass(ExplicitComponent):
    def setup(self):

        # variables
        self.add_input('rotor_mass', 0.0, units='kg', desc='mass of the rotor')
        self.add_input('nacelle_mass', 0.0, units='kg', desc='mass of nacelle')

        self.add_input('hub_system_cm', np.zeros(3), units='m', desc='location of hub center of mass relative to tower top in yaw-aligned c.s.')
        self.add_input('nacelle_cm', np.zeros(3), units='m', desc='location of nacelle center of mass relative to tower top in yaw-aligned c.s.')

        # order for all moments of inertia is (xx, yy, zz, xy, xz, yz) in the yaw-aligned coorinate system
        self.add_input('blades_I', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of all blades about hub center')
        self.add_input('hub_system_I', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of hub about its center of mass')
        self.add_input('nacelle_I', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of nacelle about its center of mass')

        # outputs
        self.add_output('rna_mass', 0.0, units='kg', desc='total mass of RNA')
        self.add_output('rna_cm', np.zeros(3), units='m', desc='location of RNA center of mass relative to tower top in yaw-aligned c.s.')
        self.add_output('rna_I_TT', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of RNA about tower top in yaw-aligned coordinate system')

        self.declare_partials('*','*')

    def _assembleI(self, I):
        Ixx, Iyy, Izz, Ixy, Ixz, Iyz = I[0], I[1], I[2], I[3], I[4], I[5] 
        return np.array([[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]])


    def _unassembleI(self, I):
        return np.array([I[0, 0], I[1, 1], I[2, 2], I[0, 1], I[0, 2], I[1, 2]])


    def compute(self, inputs, outputs):

        nac_mass = inputs['nacelle_mass']
        rotor_mass = inputs['rotor_mass']
        outputs['rna_mass'] = rotor_mass + nac_mass
        # rna cm
        outputs['rna_cm'] = (rotor_mass*inputs['hub_system_cm'] + nac_mass*inputs['nacelle_cm'])/outputs['rna_mass']

        # rna I
        blades_I = self._assembleI(inputs['blades_I'])
        hub_I = self._assembleI(inputs['hub_system_I'])
        nac_I = self._assembleI(inputs['nacelle_I'])
        rotor_I = blades_I + hub_I

        R = inputs['hub_system_cm']
        rotor_I_TT = rotor_I + rotor_mass*(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        R = inputs['nacelle_cm']
        nac_I_TT = nac_I + nac_mass*(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        outputs['rna_I_TT'] = self._unassembleI(rotor_I_TT + nac_I_TT)


    def compute_partials(self, inputs, J):

        nac_mass = inputs['nacelle_mass']
        hub_cm = inputs['hub_system_cm']
        nac_cm = inputs['nacelle_cm']
        hub_I = inputs['hub_system_I']
        nac_I = inputs['nacelle_I']
        rotor_mass = inputs['rotor_mass']
        rna_mass = rotor_mass + nac_mass
        

        # mass
        J['rna_mass', 'rotor_mass'] = 1.0
        J['rna_mass', 'nacelle_mass'] = 1.0
        J['rna_mass', 'hub_system_cm'] = np.zeros(3)
        J['rna_mass', 'nacelle_cm'] = np.zeros(3)
        J['rna_mass', 'blades_I'] = np.zeros(6)
        J['rna_mass', 'hub_system_I'] = np.zeros(6)
        J['rna_mass', 'nacelle_I'] = np.zeros(6)
        

        # cm
        numerator = rotor_mass*hub_cm+nac_mass*nac_cm

        J['rna_cm', 'rotor_mass'] = (rna_mass*hub_cm-numerator)/rna_mass**2
        J['rna_cm', 'nacelle_mass'] = (rna_mass*nac_cm-numerator)/rna_mass**2
        J['rna_cm', 'hub_system_cm'] = rotor_mass/rna_mass*np.eye(3)
        J['rna_cm', 'nacelle_cm'] = nac_mass/rna_mass*np.eye(3)
        J['rna_cm', 'blades_I'] = np.zeros((3, 6))
        J['rna_cm', 'hub_system_I'] = np.zeros((3, 6))
        J['rna_cm', 'nacelle_I'] = np.zeros((3, 6))


        # I
        R = hub_cm
        const = self._unassembleI(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        J['rna_I_TT', 'rotor_mass'] = const
        dI_drx = rotor_mass*self._unassembleI(2*R[0]*np.eye(3) - np.array([[2*R[0], R[1], R[2]], [R[1], 0.0, 0.0], [R[2], 0.0, 0.0]]))
        dI_dry = rotor_mass*self._unassembleI(2*R[1]*np.eye(3) - np.array([[0.0, R[0], 0.0], [R[0], 2*R[1], R[2]], [0.0, R[2], 0.0]]))
        dI_drz = rotor_mass*self._unassembleI(2*R[2]*np.eye(3) - np.array([[0.0, 0.0, R[0]], [0.0, 0.0, R[1]], [R[0], R[1], 2*R[2]]]))
        J['rna_I_TT', 'hub_system_cm'] = np.vstack([dI_drx, dI_dry, dI_drz]).T

        R = nac_cm
        const = self._unassembleI(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        J['rna_I_TT', 'nacelle_mass'] = const
        dI_drx = nac_mass*self._unassembleI(2*R[0]*np.eye(3) - np.array([[2*R[0], R[1], R[2]], [R[1], 0.0, 0.0], [R[2], 0.0, 0.0]]))
        dI_dry = nac_mass*self._unassembleI(2*R[1]*np.eye(3) - np.array([[0.0, R[0], 0.0], [R[0], 2*R[1], R[2]], [0.0, R[2], 0.0]]))
        dI_drz = nac_mass*self._unassembleI(2*R[2]*np.eye(3) - np.array([[0.0, 0.0, R[0]], [0.0, 0.0, R[1]], [R[0], R[1], 2*R[2]]]))
        J['rna_I_TT', 'nacelle_cm'] = np.vstack([dI_drx, dI_dry, dI_drz]).T

        J['rna_I_TT', 'blades_I'] = np.eye(6)
        J['rna_I_TT', 'hub_system_I'] = np.eye(6)
        J['rna_I_TT', 'nacelle_I'] = np.eye(6)

        

#--------------------------------------------
class RotorLoads(ExplicitComponent):
    def setup(self):

        # variables
        self.add_input('Fxyz', np.zeros(3), units='N', desc='forces in hub-aligned coordinate system')
        self.add_input('Mxyz', np.zeros(3), units='N*m', desc='moments in hub-aligned coordinate system')
        self.add_input('hub_system_cm', np.zeros(3), units='m', desc='position of rotor hub relative to tower top in yaw-aligned c.s.')
        self.add_input('rna_mass', 0.0, units='kg', desc='mass of rotor nacelle assembly')
        self.add_input('rna_cm', np.zeros(3), units='m', desc='location of RNA center of mass relative to tower top in yaw-aligned c.s.')

        self.add_discrete_input('rna_weightM', True, desc='Flag to indicate whether or not the RNA weight should be considered.\
                          An upwind overhang may lead to unconservative estimates due to the P-Delta effect(suggest not using). For downwind turbines set to True. ')

        # parameters
        self.add_discrete_input('downwind', False)
        self.add_input('tilt', 0.0, units='deg')

        # out
        self.add_output('top_F', np.zeros(3), units='N')  # in yaw-aligned
        self.add_output('top_M', np.zeros(3), units='N*m')

        self.declare_partials('top_F', ['Fxyz','Mxyz','hub_system_cm','rna_mass','rna_cm'])
        self.declare_partials('top_M', ['Fxyz','Mxyz','hub_system_cm','rna_mass','rna_cm'])


    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        F = inputs['Fxyz']
        M = inputs['Mxyz']
        tilt = float(inputs['tilt'])
        
        F = DirectionVector.fromArray(F).hubToYaw(tilt)
        M = DirectionVector.fromArray(M).hubToYaw(tilt)

        # change x-direction if downwind
        hub_cm = np.copy(inputs['hub_system_cm'])
        rna_cm = np.copy(inputs['rna_cm'])
        if discrete_inputs['downwind']:
            hub_cm[0] *= -1
            rna_cm[0] *= -1
        hub_cm = DirectionVector.fromArray(hub_cm)
        rna_cm = DirectionVector.fromArray(rna_cm)
        self.save_rhub = hub_cm
        self.save_rcm = rna_cm

        # aerodynamic moments
        M = M + hub_cm.cross(F)
        self.saveF = F


        # add weight loads
        F_w = DirectionVector(0.0, 0.0, -float(inputs['rna_mass'])*gravity)
        M_w = rna_cm.cross(F_w)
        self.saveF_w = F_w

        Fout = F + F_w

        if discrete_inputs['rna_weightM']:
            Mout = M + M_w
        else:
            Mout = M
            #REMOVE WEIGHT EFFECT TO ACCOUNT FOR P-Delta Effect
            print("!!!! No weight effect on rotor moments -TowerSE  !!!!")

        # put back in array
        outputs['top_F'] = np.array([Fout.x, Fout.y, Fout.z])
        outputs['top_M'] = np.array([Mout.x, Mout.y, Mout.z])

    def compute_partials(self, inputs, J, discrete_inputs):

        dF = DirectionVector.fromArray(inputs['Fxyz']).hubToYaw(inputs['tilt'])
        dFx, dFy, dFz = dF.dx, dF.dy, dF.dz

        dtopF_dFx = np.array([dFx['dx'], dFy['dx'], dFz['dx']])
        dtopF_dFy = np.array([dFx['dy'], dFy['dy'], dFz['dy']])
        dtopF_dFz = np.array([dFx['dz'], dFy['dz'], dFz['dz']])
        dtopF_dF = hstack([dtopF_dFx, dtopF_dFy, dtopF_dFz])
        dtopF_w_dm = np.array([0.0, 0.0, -gravity])

        #dtopF = hstack([dtopF_dF, np.zeros((3, 6)), dtopF_w_dm, np.zeros((3, 3))])


        dM = DirectionVector.fromArray(inputs['Mxyz']).hubToYaw(inputs['tilt'])
        dMx, dMy, dMz = dM.dx, dM.dy, dM.dz
        dMxcross, dMycross, dMzcross = self.save_rhub.cross_deriv(self.saveF, 'dr', 'dF')

        dtopM_dMx = np.array([dMx['dx'], dMy['dx'], dMz['dx']])
        dtopM_dMy = np.array([dMx['dy'], dMy['dy'], dMz['dy']])
        dtopM_dMz = np.array([dMx['dz'], dMy['dz'], dMz['dz']])
        dtopM_dM = hstack([dtopM_dMx, dtopM_dMy, dtopM_dMz])
        dM_dF = np.array([dMxcross['dF'], dMycross['dF'], dMzcross['dF']])

        dtopM_dFx = np.dot(dM_dF, dtopF_dFx)
        dtopM_dFy = np.dot(dM_dF, dtopF_dFy)
        dtopM_dFz = np.dot(dM_dF, dtopF_dFz)
        dtopM_dF = hstack([dtopM_dFx, dtopM_dFy, dtopM_dFz])
        dtopM_dr = np.array([dMxcross['dr'], dMycross['dr'], dMzcross['dr']])

        dMx_w_cross, dMy_w_cross, dMz_w_cross = self.save_rcm.cross_deriv(self.saveF_w, 'dr', 'dF')

        if discrete_inputs['rna_weightM']:
            dtopM_drnacm = np.array([dMx_w_cross['dr'], dMy_w_cross['dr'], dMz_w_cross['dr']])
            dtopM_dF_w = np.array([dMx_w_cross['dF'], dMy_w_cross['dF'], dMz_w_cross['dF']])
        else:
            dtopM_drnacm = np.zeros((3, 3))
            dtopM_dF_w = np.zeros((3, 3))
        dtopM_dm = np.dot(dtopM_dF_w, dtopF_w_dm)

        if discrete_inputs['downwind']:
            dtopM_dr[:, 0] *= -1
            dtopM_drnacm[:, 0] *= -1

        #dtopM = hstack([dtopM_dF, dtopM_dM, dtopM_dr, dtopM_dm, dtopM_drnacm])

        
        J['top_F', 'Fxyz'] = dtopF_dF
        J['top_F', 'Mxyz'] = np.zeros((3, 3))
        J['top_F', 'hub_system_cm'] = np.zeros((3, 3))
        J['top_F', 'rna_mass'] = dtopF_w_dm
        J['top_F', 'rna_cm'] = np.zeros((3, 3))

        J['top_M', 'Fxyz'] = dtopM_dF
        J['top_M', 'Mxyz'] = dtopM_dM
        J['top_M', 'hub_system_cm'] = dtopM_dr
        J['top_M', 'rna_mass'] = dtopM_dm
        J['top_M', 'rna_cm'] = dtopM_drnacm

        

#-------------------------------------------------------------------------
# Groups
#   (were Assemblies in OpenMDAO 0.x)
#------------------------------------------------------------------

class DriveSE(Group):
    ''' Class Drive4pt defines an OpenMDAO group that represents a wind turbine drivetrain with a 4-point suspension
      (two main bearings). This Group can serve as the root of an OpenMDAO Problem.
    '''

    def initialize(self):
        self.options.declare('number_of_main_bearings')
        self.options.declare('topLevelFlag', default=True)
        self.options.declare('debug', default=False)
        
    def setup(self):
        debug=self.options['debug']
        if not self.options['number_of_main_bearings'] in [1,2]:
            raise ValueError('Number of main bearings must be one or two')
        elif self.options['number_of_main_bearings'] == 2:
            drive4pt = True
        else:
            drive4pt = False

        # Independent variables that are unique to DriveSE
        driveIndeps = IndepVarComp()
        driveIndeps.add_output('gear_ratio', 0.0)
        driveIndeps.add_output('shaft_angle', 0.0, units='rad')
        driveIndeps.add_output('shaft_ratio', 0.0)
        driveIndeps.add_output('shrink_disc_mass', 0.0, units='kg')
        driveIndeps.add_output('carrier_mass', 0.0, units='kg')
        driveIndeps.add_output('flange_length', 0.0, units='m')
        driveIndeps.add_output('overhang', 0.0, units='m')
        driveIndeps.add_output('distance_hub2mb', 0.0, units='m')
        driveIndeps.add_output('gearbox_input_xcm', 0.0, units='m')
        driveIndeps.add_output('hss_input_length', 0.0, units='m')
        driveIndeps.add_discrete_output('planet_numbers', np.array([0, 0, 0]))
        driveIndeps.add_discrete_output('drivetrain_design', 'geared')
        driveIndeps.add_discrete_output('gear_configuration', 'eep')
        if drive4pt:
            driveIndeps.add_discrete_output('mb1Type', 'CARB')
            driveIndeps.add_discrete_output('mb2Type', 'SRB')
        else:
            driveIndeps.add_discrete_output('mb1Type', 'SRB')
        driveIndeps.add_discrete_output('IEC_Class', 'B')
        driveIndeps.add_discrete_output('shaft_factor', 'normal')
        driveIndeps.add_discrete_output('uptower_transformer', True)
        driveIndeps.add_discrete_output('crane', True)
        driveIndeps.add_discrete_output('rna_weightM', True)
        driveIndeps.add_discrete_output('downwind', True)
        driveIndeps.add_discrete_output('yaw_motors_number', 0)
        self.add_subsystem('driveIndeps', driveIndeps, promotes=['*'])

        # Independent variables that may be duplicated at higher levels of aggregation
        if self.options['topLevelFlag']:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_discrete_output('number_of_blades', 0)
            sharedIndeps.add_output('tower_top_diameter',     0.0, units='m')
            sharedIndeps.add_output('rotor_diameter',         0.0, units='m')
            sharedIndeps.add_output('rotor_rpm',              0.0, units='rpm')
            sharedIndeps.add_output('rotor_torque',           0.0, units='N*m')
            sharedIndeps.add_output('Fxyz',                   np.zeros(3), units='N')
            sharedIndeps.add_output('Mxyz',                   np.zeros(3), units='N*m')
            sharedIndeps.add_output('blade_mass',             0.0, units='kg')
            sharedIndeps.add_output('blade_root_diameter',    0.0, units='m')
            sharedIndeps.add_output('blade_length',           0.0, units='m')
            sharedIndeps.add_output('blades_I',               np.zeros(6), units='kg*m**2')
            sharedIndeps.add_output('drivetrain_efficiency',  0.0)
            sharedIndeps.add_output('tile',                   0.0, units='deg')
            sharedIndeps.add_output('machine_rating',         0.0, units='kW')
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])

        # select components
        self.add_subsystem('demux', ForceMomentDemux(), promotes=['*'])
        self.add_subsystem('hub', HubSE(mass_only=True, topLevelFlag=False, debug=debug), promotes=['*'])
        self.add_subsystem('gearbox', Gearbox_OM(), promotes=['*'])
        self.add_subsystem('lowSpeedShaft', LowSpeedShaft4pt_OM(), promotes=['*'])
        self.add_subsystem('mainBearing', MainBearing_OM(bearing_position='main'), promotes=['lss_design_torque','rotor_diameter']) #explicit connections for bearings
        if drive4pt:
            self.add_subsystem('secondBearing', MainBearing_OM(bearing_position='second'), promotes=['lss_design_torque','rotor_diameter']) #explicit connections for bearings
        self.add_subsystem('hubCM', Hub_CM_Adder_OM(), promotes=['*'])
        
        self.add_subsystem('highSpeedSide', HighSpeedSide_OM(), promotes=['*'])
        self.add_subsystem('generator', Generator_OM(), promotes=['*'])
        self.add_subsystem('transformer', Transformer_OM(), promotes=['*'])
        self.add_subsystem('bedplate', Bedplate_OM(), promotes=['*'])
        self.add_subsystem('above_yaw_massAdder', AboveYawMassAdder_OM(), promotes=['*'])
        self.add_subsystem('yawSystem', YawSystem_OM(), promotes=['*'])
        self.add_subsystem('nacelleSystem', NacelleSystemAdder_OM(), promotes=['*'])
        self.add_subsystem('rna', RNAMass(), promotes=['*'])
        self.add_subsystem('loads', RotorLoads(), promotes=['*'])

        # Connect components where explicit connections needed (for main bearings)
        self.connect('lss_mb1_mass',        ['mainBearing.bearing_mass'])
        self.connect('lss_diameter1',       ['mainBearing.lss_diameter', 'lss_diameter'])
        self.connect('lss_mb1_cm',          ['mainBearing.lss_mb_cm'])
        self.connect('mainBearing.mb_mass', ['mb1_mass'])
        self.connect('mainBearing.mb_cm',   ['mb1_cm', 'MB1_location'])
        self.connect('mainBearing.mb_I',    ['mb1_I'])

        if drive4pt:
            self.connect('lss_mb2_mass',          ['secondBearing.bearing_mass'])
            self.connect('lss_diameter2',         ['secondBearing.lss_diameter'])
            self.connect('lss_mb2_cm',            ['secondBearing.lss_mb_cm'])
            self.connect('secondBearing.mb_mass', ['mb2_mass'])
            self.connect('secondBearing.mb_cm',   ['mb2_cm'])
            self.connect('secondBearing.mb_I',    ['mb2_I'])

        self.connect('lss_cm',       'lss_location',       src_indices=[0])
        self.connect('hss_cm',       'hss_location',       src_indices=[0])
        self.connect('gearbox_cm',   'gearbox_location',   src_indices=[0])
        self.connect('generator_cm', 'generator_location', src_indices=[0])
            

#------------------------------------------------------------------
# examples

def nacelle_example_5MW_baseline_3pt(debug=False):

    # NREL 5 MW Drivetrain variables
    # geared 3-stage Gearbox with induction generator machine

    runid = 'N5_3pt'
    modid = ''
    
    prob=Problem()
    prob.model=DriveSE(debug=debug, number_of_main_bearings=1, topLevelFlag=True)
    prob.setup()
    #view_connections(prob.model, show_browser=True)

    prob['drivetrain_design']='geared'
    prob['gear_configuration']='eep'  # epicyclic-epicyclic-parallel
    prob['mb1Type']='SRB'
    prob['IEC_Class']='B'
    prob['shaft_factor']='normal'
    prob['uptower_transformer']=True
    prob['crane']=True  # onboard crane present
    prob['yaw_motors_number'] = 0 # default value - will be internally calculated
    prob['number_of_blades'] = 3
    
    # Rotor and load inputs
    prob['rotor_diameter'] = 126.0  # m
    prob['rotor_rpm'] = 12.1  # rpm m/s
    prob['tilt'] = 5.0 # deg
    prob['machine_rating'] = 5000.0
    prob['drivetrain_efficiency'] = 0.95
    prob['rotor_torque'] = 1.5 * (prob['machine_rating'] * 1000 / prob['drivetrain_efficiency']) \
                              / (prob['rotor_rpm'] * (np.pi / 30))
    #prob['rotor_thrust'] = 599610.0  # N
    prob['rotor_mass'] = 0.0  # accounted for in F_z # kg
    prob['Mxyz'] = np.array([330770.0, -16665000.0, 2896300.0])  # Nm
    prob['Fxyz'] = np.array([ 599610.0, 186780.0, -842710.0])  # N

    # Drivetrain inputs
    prob['machine_rating'] = 5000.0  # kW
    prob['gear_ratio'] = 96.76  # 97:1 as listed in the 5 MW reference document
    prob['shaft_angle'] = 5.0*np.pi / 180.0  # rad
    prob['shaft_ratio'] = 0.10 # 0.10 may be a bit small!
    prob['planet_numbers'] = [3, 3, 1]
    prob['shrink_disc_mass'] = 333.3 * prob['machine_rating'] / 1000.0  # estimated
    prob['carrier_mass'] = 8000.0  # estimated
    #prob['flange_length'] = 0.5
    prob['overhang'] = 5.0
    prob['distance_hub2mb'] = 1.912  # length from hub center to main bearing, leave zero if unknown
    prob['gearbox_input_xcm'] = 0.1
    prob['hss_input_length'] = 1.5

    # NREL 5MW properties from RotorSE
    prob['blade_mass'] = 16539.70885528
    prob['blade_root_diameter'] = 2.5
    prob['blade_length'] = 60.0
    prob['blades_I'] = np.array([37125777.37275547, 17852835.54452561, 14672945.83129043, 0.0, 0.0, 0.0])
    
    
    prob['hub_flange_thickness'] = 0
    
    # Tower inputs
    prob['tower_top_diameter'] = 3.78  # m

    # test cases
    #prob['rotor_mass'] = 1000; modid = '_r1k'
    
    prob.run_driver()
    prob.model.list_inputs(units=True)#values = False, hierarchical=False)
    prob.model.list_outputs(units=True)#values = False, hierarchical=False)    

    return prob

#-------------------------------------------------------------------------

def nacelle_example_5MW_baseline_4pt(debug=False):

    # NREL 5 MW Drivetrain variables
    # geared 3-stage Gearbox with induction generator machine
    prob=Problem()
    prob.model=DriveSE(debug=debug, number_of_main_bearings=2, topLevelFlag=True)
    prob.setup()

    prob['drivetrain_design']='geared'
    prob['gear_configuration']='eep'  # epicyclic-epicyclic-parallel
    prob['mb1Type']='CARB'
    prob['mb2Type']='SRB'
    prob['IEC_Class']='B'
    prob['shaft_factor']='normal'
    prob['uptower_transformer']=True
    prob['crane']=True  # onboard crane present
    prob['yaw_motors_number'] = 0 # default value - will be internally calculated
    prob['number_of_blades'] = 3
    
    # Rotor and load inputs
    prob['rotor_diameter'] = 126.0  # m
    prob['rotor_rpm'] = 12.1  # rpm m/s
    prob['tilt'] = 5.0 # deg
    prob['machine_rating'] = 5000.0
    prob['drivetrain_efficiency'] = 0.95
    prob['rotor_torque'] = 1.5 * (prob['machine_rating'] * 1000 / \
                             prob['drivetrain_efficiency']) / (prob['rotor_rpm'] * (np.pi / 30))
    prob['rotor_mass'] = 0.0  # accounted for in F_z # kg
    prob['Mxyz'] = np.array([330770.0, -16665000.0, 2896300.0])  # Nm
    prob['Fxyz'] = np.array([ 599610.0, 186780.0, -842710.0])  # N

    # Drivetrain inputs
    prob['machine_rating'] = 5000.0  # kW
    prob['gear_ratio'] = 96.76  # 97:1 as listed in the 5 MW reference document
    prob['shaft_angle'] = 5.0*np.pi / 180.0  # rad
    prob['shaft_ratio'] = 0.10
    prob['planet_numbers'] = [3, 3, 1]
    prob['shrink_disc_mass'] = 333.3 * prob['machine_rating'] / 1000.0  # estimated
    prob['carrier_mass'] = 8000.0  # estimated
    prob['flange_length'] = 0.5
    prob['overhang'] = 5.0
    prob['distance_hub2mb'] = 1.912  # length from hub center to main bearing, leave zero if unknown
    prob['gearbox_input_xcm'] = 0.1
    prob['hss_input_length'] = 1.5

    # NREL 5MW properties from RotorSE
    prob['blade_mass'] = 16539.70885528
    prob['blade_root_diameter'] = 2.5
    prob['blade_length'] = 60.0
    prob['blades_I'] = np.array([37125777.37275547, 17852835.54452561, 14672945.83129043, 0.0, 0.0, 0.0])
    
    # Tower inputs
    prob['tower_top_diameter'] = 3.78  # m

    prob.run_driver()
    prob.model.list_inputs(units=True)#values = False, hierarchical=False)
    prob.model.list_outputs(units=True)#values = False, hierarchical=False)    

    return prob

'''
#Need to update for new structure of drivetrain
def nacelle_example_1p5MW_3pt():

    # test of module for turbine data set

    # 1.5 MW Rotor Variables
    print('----- NREL 1p5MW  Drivetrain - 3 Point Suspension-----')
    nace=Group()
    Drive3pt(nace)
    prob=Problem(nace)
    prob.setup()
    nace.rotor_diameter=77  # m
    nace.rotor_rpm=16.18  # rpm
    nace.drivetrain_efficiency=0.95
    nace.machine_rating=1500
    nace.rotor_torque=1.5 * (nace.machine_rating * 1000 / nace.drivetrain_efficiency) / \
                             (nace.rotor_rpm * (pi / 30)
                              )  # 6.35e6 #4365248.74 # Nm
    nace.rotor_thrust=2.6204e5
    nace.rotor_mass=0.0
    nace.rotor_bending_moment=2.7795e6
    nace.rotor_bending_moment_x=8.4389e5
    nace.rotor_bending_moment_y=-2.6758e6
    nace.rotor_bending_moment_z=7.5222e2
    nace.rotor_thrust=2.6204e5
    nace.rotor_force_y=2.8026e4
    nace.rotor_force_z=-3.4763e5


    # 1p5MW  Drivetrain variables
    # geared 3-stage Gearbox with induction generator machine
    nace.drivetrain_design='geared'
    nace.machine_rating=1500.0  # kW
    nace.gear_ratio=78
    nace.gear_configuration='epp'  # epicyclic-parallel-parallel
    nace.crane=False  # onboard crane not present
    nace.shaft_angle=5.0  # deg
    nace.shaft_ratio=0.10
    nace.Np=[3, 1, 1]
    nace.shaft_type='normal'
    nace.uptower_transformer=False  # True
    nace.shrink_disc_mass=333.3 * nace.machine_rating / 1000.0  # estimated
    nace.carrier_mass=2000.0  # estimated
    nace.mb1Type='SRB'
    nace.mb2Type='SRB'
    nace.flange_length=0.285  # m
    nace.overhang=3.3
    nace.distance_hub2mb=1.535  # length from hub center to main bearing, leave zero if unknown

    # 0 if no fatigue check, 1 if parameterized fatigue check, 2 if known
    # loads inputs
    nace.check_fatigue=0

    # variables if check_fatigue = 1:
    nace.blade_number=3
    nace.cut_in=3.5  # cut-in m/s
    nace.cut_out=20.  # cut-out m/s
    nace.Vrated=11.5  # rated windspeed m/s
    nace.weibull_k=2.2  # windepeed distribution shape parameter
    nace.weibull_A=9.  # windspeed distribution scale parameter
    nace.T_life=20.  # design life in years
    nace.IEC_Class_Letter='B'

    # variables if check_fatigue =2:
    # nace.rotor_thrust_distribution =
    # nace.rotor_thrust_count =
    # nace.rotor_Fy_distribution =
    # nace.rotor_Fy_count =
    # nace.rotor_Fz_distribution =
    # nace.rotor_Fz_count =
    # nace.rotor_torque_distribution =
    # nace.rotor_torque_count =
    # nace.rotor_My_distribution =
    # nace.rotor_My_count =
    # nace.rotor_Mz_distribution =
    # nace.rotor_Mz_count =

    # 1p5MW Tower Variables
    nace.tower_top_diameter=2.3  # m

    prob.run_driver()

    sys_print(nace)

def nacelle_example_1p5MW_4pt():

    # test of module for turbine data set

    print('----- NREL 1p5MW  Drivetrain - 4 Point Suspension-----')
    nace=Group()
    Drive4pt(nace)
    prob=Problem(nace)
    prob.setup()
    nace.rotor_diameter=77  # m
    nace.rotor_rpm=16.18  # rpm
    nace.drivetrain_efficiency=0.95
    nace.machine_rating=1500
    nace.rotor_torque=1.5 * (nace.machine_rating * 1000 / nace.drivetrain_efficiency) / \
                             (nace.rotor_rpm * (pi / 30)
                              )  # 6.35e6 #4365248.74 # Nm
    nace.rotor_thrust=2.6204e5
    nace.rotor_mass=0.0
    nace.rotor_bending_moment=2.7795e6
    nace.rotor_bending_moment_x=8.4389e5
    nace.rotor_bending_moment_y=-2.6758e6
    nace.rotor_bending_moment_z=7.5222e2
    nace.rotor_thrust=2.6204e5
    nace.rotor_force_y=2.8026e4
    nace.rotor_force_z=-3.4763e5

    # 1p5MW  Drivetrain variables
    # geared 3-stage Gearbox with induction generator machine
    nace.drivetrain_design='geared'
    nace.machine_rating=1500.0  # kW
    nace.gear_ratio=78
    nace.gear_configuration='epp'  # epicyclic-parallel-parallel
    nace.crane=False  # True # onboard crane present
    nace.shaft_angle=5.0  # deg
    nace.shaft_ratio=0.10
    nace.Np=[3, 1, 1]
    nace.shaft_type='normal'
    nace.uptower_transformer=False  # True
    nace.shrink_disc_mass=333.3 * nace.machine_rating / 1000.0  # estimated
    nace.carrier_mass=2000.0  # estimated
    nace.mb1Type='CARB'
    nace.mb2Type='SRB'
    nace.flange_length=0.285  # m
    nace.overhang=4
    nace.distance_hub2mb=1.3  # length from hub center to main bearing, leave zero if unknown
    nace.gearbox_cm=0.0

    # 0 if no fatigue check, 1 if parameterized fatigue check, 2 if known
    # loads inputs
    nace.check_fatigue=0

    # variables if check_fatigue = 1:
    nace.blade_number=3
    nace.cut_in=3.5  # cut-in m/s
    nace.cut_out=20.  # cut-out m/s
    nace.Vrated=11.5  # rated windspeed m/s
    nace.weibull_k=2.2  # windepeed distribution shape parameter
    nace.weibull_A=9.  # windspeed distribution scale parameter
    nace.T_life=20.  # design life in years
    nace.IEC_Class_Letter='B'

    # variables if check_fatigue =2:
    # nace.rotor_thrust_distribution =
    # nace.rotor_thrust_count =
    # nace.rotor_Fy_distribution =
    # nace.rotor_Fy_count =
    # nace.rotor_Fz_distribution =
    # nace.rotor_Fz_count =
    # nace.rotor_torque_distribution =
    # nace.rotor_torque_count =
    # nace.rotor_My_distribution =
    # nace.rotor_My_count =
    # nace.rotor_Mz_distribution =
    # nace.rotor_Mz_count =

    # 1p5MW Tower Variables
    nace.tower_top_diameter=2.3  # m

    prob.run_driver()

    # cm_print(nace)
    sys_print(nace)

def nacelle_example_p75_3pt():

    # test of module for turbine data set
    print('----- NREL 750kW Design - 3 Point Suspension----')
    # 0.75MW Rotor Variables
    nace=Group()
    Drive3pt(nace)
    prob=Problem(nace)
    prob.setup()
    nace.rotor_diameter=48.2  # m
    nace.rotor_rpm=22.0  # rpm m/s
    nace.drivetrain_efficiency=0.95
    nace.machine_rating=750
    nace.rotor_torque=1.5 * (nace.machine_rating * 1000 / nace.drivetrain_efficiency) / \
                             (nace.rotor_rpm * (pi / 30)
                              )  # 6.35e6 #4365248.74 # Nm

    nace.rotor_thrust=143000.0  # N
    nace.rotor_mass=0.0  # kg
    nace.rotor_bending_moment=495.6e3
    nace.rotor_bending_moment_x=401.0e3
    nace.rotor_bending_moment_y=495.6e3
    nace.rotor_bending_moment_z=-443.0e3
    nace.rotor_thrust=143000.0
    nace.rotor_force_y=-12600.0
    nace.rotor_force_z=-142.0e3

    # NREL 750 kW Drivetrain variables
    # geared 3-stage Gearbox with induction generator machine
    nace.drivetrain_design='geared'
    nace.machine_rating=750  # kW
    nace.gear_ratio=81.491
    nace.gear_configuration='epp'  # epicyclic-parallel-parallel
    nace.crane=False  # True if onboard crane present
    nace.shaft_angle=5.0  # deg
    nace.shaft_length=2.1  # m
    nace.shaft_ratio=0.10
    nace.Np=[3, 1, 1]
    nace.shaft_type='normal'
    nace.uptower_transformer=False
    nace.shrink_disc_mass=333.3 * nace.machine_rating / 1000.0  # estimated
    nace.carrier_mass=250.  # estimated
    nace.mb1Type='SRB'
    nace.mb2Type='TRB2'
    nace.flange_length=0.285  # m
    nace.overhang=2.26
    nace.distance_hub2mb=1.22  # length from hub center to main bearing, leave zero if unknown
    nace.gearbox_cm=0.8
    nace.blade_root_diameter=1.6

    # 0 if no fatigue check, 1 if parameterized fatigue check, 2 if known
    # loads inputs
    nace.check_fatigue=0

    # variables if check_fatigue = 1:
    nace.blade_number=3
    nace.cut_in=3.  # cut-in m/s
    nace.cut_out=25.  # cut-out m/s
    nace.Vrated=16.  # rated windspeed m/s
    nace.weibull_k=2.2  # windepeed distribution shape parameter
    nace.weibull_A=9.  # windspeed distribution scale parameter
    nace.T_life=20.  # design life in years
    nace.IEC_Class_Letter='A'


    # variables if check_fatigue =2:
    # nace.rotor_thrust_distribution =
    # nace.rotor_thrust_count =
    # nace.rotor_Fy_distribution =
    # nace.rotor_Fy_count =
    # nace.rotor_Fz_distribution =
    # nace.rotor_Fz_count =
    # nace.rotor_torque_distribution =
    # nace.rotor_torque_count =
    # nace.rotor_My_distribution =
    # nace.rotor_My_count =
    # nace.rotor_Mz_distribution =
    # nace.rotor_Mz_count =

    # 0.75MW Tower Variables
    nace.tower_top_diameter=2.21  # m

    prob.run_driver()
    # cm_print(nace)
    sys_print(nace)

def nacelle_example_p75_4pt():

    # test of module for turbine data set
    print('----- NREL 750kW Design - 4 Point Suspension----')
    # 0.75MW Rotor Variables
    nace=Group()
    Drive4pt(nace)
    prob=Problem(nace)
    prob.setup()
    nace.rotor_diameter=48.2  # m
    nace.rotor_rpm=22.0  # rpm
    nace.drivetrain_efficiency=0.95
    nace.machine_rating=750
    nace.rotor_torque=1.5 * (nace.machine_rating * 1000 / nace.drivetrain_efficiency) / \
                             (nace.rotor_rpm * (pi / 30)
                              )  # 6.35e6 #4365248.74 # Nm
    #nace.rotor_torque = 6.37e6 #
    nace.rotor_thrust=143000.0
    nace.rotor_mass=0.0
    nace.rotor_bending_moment=459.6e3
    nace.rotor_bending_moment_x=401.0e3
    nace.rotor_bending_moment_y=459.6e3
    nace.rotor_bending_moment_z=-443.0e3
    nace.rotor_thrust=143000.0
    nace.rotor_force_y=-12600.0
    nace.rotor_force_z=-142.0e3

    # NREL 750 kW Drivetrain variables
    # geared 3-stage Gearbox with induction generator machine
    nace.drivetrain_design='geared'
    nace.machine_rating=750  # kW
    nace.gear_ratio=81.491  # as listed in the 5 MW reference document
    nace.gear_configuration='epp'  # epicyclic-parallel-parallel
    nace.crane=False  # True # onboard crane present
    nace.shaft_angle=5.0  # deg
    nace.shaft_length=2.1  # m
    nace.shaft_ratio=0.10
    nace.Np=[3, 1, 1]
    nace.shaft_type='normal'
    nace.uptower_transformer=False  # True
    nace.shrink_disc_mass=333.3 * nace.machine_rating / 1000.0  # estimated
    nace.carrier_mass=1000.0  # estimated
    nace.mb1Type='SRB'
    nace.mb2Type='TRB2'
    nace.flange_length=0.338  # m
    nace.overhang=2.26
    nace.distance_hub2mb=1.22  # 0.007835*rotor_diameter+0.9642 length from hub center to main bearing, leave zero if unknown
    nace.gearbox_cm=0.90

    # 0 if no fatigue check, 1 if parameterized fatigue check, 2 if known
    # loads inputs
    nace.check_fatigue=0

    # variables if check_fatigue = 1:
    nace.blade_number=3
    nace.cut_in=3.  # cut-in m/s
    nace.cut_out=25.  # cut-out m/s
    nace.Vrated=16.  # rated windspeed m/s
    nace.weibull_k=2.2  # windepeed distribution shape parameter
    nace.weibull_A=9.  # windspeed distribution scale parameter
    nace.T_life=20.  # design life in years
    nace.IEC_Class_Letter='A'

    # variables if check_fatigue =2:
    # nace.rotor_thrust_distribution =
    # nace.rotor_thrust_count =
    # nace.rotor_Fy_distribution =
    # nace.rotor_Fy_count =
    # nace.rotor_Fz_distribution =
    # nace.rotor_Fz_count =
    # nace.rotor_torque_distribution =
    # nace.rotor_torque_count =
    # nace.rotor_My_distribution =
    # nace.rotor_My_count =
    # nace.rotor_Mz_distribution =
    # nace.rotor_Mz_count =

    # 0.75MW Tower Variables
    nace.tower_top_diameter=2.21  # m

    prob.run_driver()

    sys_print(nace)
'''

if __name__ == '__main__':
    ''' Main runs through tests of several drivetrain configurations with known component masses and dimensions '''

    debug = True
    #debug = False
    
    nacelle_example_5MW_baseline_3pt(debug=debug)

    #nacelle_example_5MW_baseline_4pt(debug=debug)
    
    '''
    nacelle_example_1p5MW_3pt()

    nacelle_example_1p5MW_4pt()

    nacelle_example_p75_3pt()

    nacelle_example_p75_4pt()'''
